# mypy: ignore-errors
import builtins
import sys
from types import ModuleType
from typing import Any

import pytest

real_import = builtins.__import__

try:
    import pystac_client  # noqa:F401

    HAS_PYSTAC_CLIENT = True
except ImportError:
    HAS_PYSTAC_CLIENT = False


@pytest.mark.skipif(not HAS_PYSTAC_CLIENT, reason="requires pystac_client")
def test_import_pystac_client_when_available() -> None:
    from pystac.client import Client

    assert Client.__doc__


# copied from
# http://materials-scientist.com/blog/2021/02/11/mocking-failing-module-import-python/
def monkey_import_importerror(name: str, *args: Any, **kwargs: Any) -> ModuleType:
    if name == "pystac_client":
        raise ImportError(f"No module named '{name}'")
    return real_import(name, *args, **kwargs)


def test_import_pystac_client_when_not_available(monkeypatch: Any) -> None:
    if HAS_PYSTAC_CLIENT:
        monkeypatch.delitem(sys.modules, "pystac_client", raising=False)
        monkeypatch.setattr(builtins, "__import__", monkey_import_importerror)

    with pytest.raises(ImportError):
        from pystac_client import Client  # noqa:F401

    with pytest.raises(ImportError, match="Please install pystac-client:"):
        from pystac.client import Client  # noqa:F401,F811



from __future__ import annotations

from collections.abc import Iterable, Sequence
from copy import deepcopy
from datetime import datetime, timezone
from typing import (
    TYPE_CHECKING,
    Any,
    Optional,
    TypeVar,
    cast,
)

import pystac
from pystac import CatalogType, STACObjectType
from pystac.asset import Asset, Assets
from pystac.catalog import Catalog
from pystac.errors import DeprecatedWarning, ExtensionNotImplemented, STACTypeError
from pystac.item_assets import ItemAssetDefinition, _ItemAssets
from pystac.layout import HrefLayoutStrategy
from pystac.link import Link
from pystac.provider import Provider
from pystac.serialization import (
    identify_stac_object,
    identify_stac_object_type,
    migrate_to_latest,
)
from pystac.summaries import Summaries
from pystac.utils import (
    datetime_to_str,
    str_to_datetime,
)

if TYPE_CHECKING:
    from pystac.extensions.ext import CollectionExt
    from pystac.item import Item

#: Generalized version of :class:`Collection`
C = TypeVar("C", bound="Collection")

Bboxes = list[list[float | int]]
TemporalIntervals = list[list[datetime]] | list[list[Optional[datetime]]]
TemporalIntervalsLike = TemporalIntervals | list[datetime] | list[Optional[datetime]]


class SpatialExtent:
    """Describes the spatial extent of a Collection.

    Args:
        bboxes : A list of bboxes that represent the spatial
            extent of the collection. Each bbox can be 2D or 3D. The length of the bbox
            array must be 2*n where n is the number of dimensions. For example, a
            2D Collection with only one bbox would be [[xmin, ymin, xmax, ymax]]

        extra_fields : Dictionary containing additional top-level fields defined on the
            Spatial Extent object.
    """

    bboxes: Bboxes
    """A list of bboxes that represent the spatial
    extent of the collection. Each bbox can be 2D or 3D. The length of the bbox
    array must be 2*n where n is the number of dimensions. For example, a
    2D Collection with only one bbox would be [[xmin, ymin, xmax, ymax]]"""

    extra_fields: dict[str, Any]
    """Dictionary containing additional top-level fields defined on the Spatial
    Extent object."""

    def __init__(
        self,
        bboxes: Bboxes | Sequence[float | int],
        extra_fields: dict[str, Any] | None = None,
    ) -> None:
        if not isinstance(bboxes, list):
            raise TypeError("bboxes must be a list")

        # A common mistake is to pass in a single bbox instead of a list of bboxes.
        # Account for this by transforming the input in that case.
        if isinstance(bboxes[0], (float, int)):
            self.bboxes = [cast(list[float | int], bboxes)]
        else:
            self.bboxes = cast(Bboxes, bboxes)

        self.extra_fields = extra_fields or {}

    def to_dict(self) -> dict[str, Any]:
        """Returns this spatial extent as a dictionary.

        Returns:
            dict: A serialization of the SpatialExtent.
        """
        d = {"bbox": self.bboxes, **self.extra_fields}
        return d

    def clone(self) -> SpatialExtent:
        """Clones this object.

        Returns:
            SpatialExtent: The clone of this object.
        """
        cls = self.__class__
        return cls(
            bboxes=deepcopy(self.bboxes), extra_fields=deepcopy(self.extra_fields)
        )

    @staticmethod
    def from_dict(d: dict[str, Any]) -> SpatialExtent:
        """Constructs a SpatialExtent from a dict.

        Returns:
            SpatialExtent: The SpatialExtent deserialized from the JSON dict.
        """
        return SpatialExtent(
            bboxes=d["bbox"], extra_fields={k: v for k, v in d.items() if k != "bbox"}
        )

    @staticmethod
    def from_coordinates(
        coordinates: list[Any], extra_fields: dict[str, Any] | None = None
    ) -> SpatialExtent:
        """Constructs a SpatialExtent from a set of coordinates.

        This method will only produce a single bbox that covers all points
        in the coordinate set.

        Args:
            coordinates : Coordinates to derive the bbox from.
            extra_fields : Dictionary containing additional top-level fields defined on
                the SpatialExtent object.

        Returns:
            SpatialExtent: A SpatialExtent with a single bbox that covers the
            given coordinates.
        """

        def process_coords(
            coord_lists: list[Any],
            xmin: float | None = None,
            ymin: float | None = None,
            xmax: float | None = None,
            ymax: float | None = None,
        ) -> tuple[float | None, float | None, float | None, float | None]:
            for coord in coord_lists:
                if isinstance(coord[0], list):
                    xmin, ymin, xmax, ymax = process_coords(
                        coord, xmin, ymin, xmax, ymax
                    )
                else:
                    x, y = coord
                    if xmin is None or x < xmin:
                        xmin = x
                    elif xmax is None or xmax < x:
                        xmax = x
                    if ymin is None or y < ymin:
                        ymin = y
                    elif ymax is None or ymax < y:
                        ymax = y
            return xmin, ymin, xmax, ymax

        xmin, ymin, xmax, ymax = process_coords(coordinates)

        if xmin is None or ymin is None or xmax is None or ymax is None:
            raise ValueError(
                f"Could not determine bounds from coordinate sequence {coordinates}"
            )

        return SpatialExtent(
            bboxes=[[xmin, ymin, xmax, ymax]], extra_fields=extra_fields
        )


class TemporalExtent:
    """Describes the temporal extent of a Collection.

    Args:
        intervals :  A list of two datetimes wrapped in a list,
            representing the temporal extent of a Collection. Open date ranges are
            supported by setting either the start (the first element of the interval)
            or the end (the second element of the interval) to None.

        extra_fields : Dictionary containing additional top-level fields defined on the
            Temporal Extent object.
    Note:
        Datetimes are required to be in UTC.
    """

    intervals: TemporalIntervals
    """A list of two datetimes wrapped in a list,
    representing the temporal extent of a Collection. Open date ranges are
    represented by either the start (the first element of the interval) or the
    end (the second element of the interval) being None."""

    extra_fields: dict[str, Any]
    """Dictionary containing additional top-level fields defined on the Temporal
    Extent object."""

    def __init__(
        self,
        intervals: TemporalIntervals | Sequence[datetime | None],
        extra_fields: dict[str, Any] | None = None,
    ):
        if not isinstance(intervals, list):
            raise TypeError("intervals must be a list")
        # A common mistake is to pass in a single interval instead of a
        # list of intervals. Account for this by transforming the input
        # in that case.
        if isinstance(intervals[0], datetime) or intervals[0] is None:
            self.intervals = [cast(list[Optional[datetime]], intervals)]
        else:
            self.intervals = cast(TemporalIntervals, intervals)

        self.extra_fields = extra_fields or {}

    def to_dict(self) -> dict[str, Any]:
        """Returns this temporal extent as a dictionary.

        Returns:
            dict: A serialization of the TemporalExtent.
        """
        encoded_intervals: list[list[str | None]] = []
        for i in self.intervals:
            start = None
            end = None

            if i[0] is not None:
                start = datetime_to_str(i[0])

            if i[1] is not None:
                end = datetime_to_str(i[1])

            encoded_intervals.append([start, end])

        d = {"interval": encoded_intervals, **self.extra_fields}
        return d

    def clone(self) -> TemporalExtent:
        """Clones this object.

        Returns:
            TemporalExtent: The clone of this object.
        """
        cls = self.__class__
        return cls(
            intervals=deepcopy(self.intervals), extra_fields=deepcopy(self.extra_fields)
        )

    @staticmethod
    def from_dict(d: dict[str, Any]) -> TemporalExtent:
        """Constructs an TemporalExtent from a dict.

        Returns:
            TemporalExtent: The TemporalExtent deserialized from the JSON dict.
        """
        parsed_intervals: list[list[datetime | None]] = []
        for i in d["interval"]:
            if isinstance(i, str):
                import warnings

                # d["interval"] is a list of strings, so we correct the list and
                # try again
                # https://github.com/stac-utils/pystac/issues/1221
                warnings.warn(
                    "A collection's temporal extent should be a list of lists, but "
                    "is instead a "
                    "list of strings. pystac is fixing this issue and continuing "
                    "deserialization, but note that the source "
                    "collection is invalid STAC.",
                    UserWarning,
                )
                d["interval"] = [d["interval"]]
                return TemporalExtent.from_dict(d)
            start = None
            end = None

            if i[0]:
                start = str_to_datetime(i[0])
            if i[1]:
                end = str_to_datetime(i[1])
            parsed_intervals.append([start, end])

        return TemporalExtent(
            intervals=parsed_intervals,
            extra_fields={k: v for k, v in d.items() if k != "interval"},
        )

    @staticmethod
    def from_now() -> TemporalExtent:
        """Constructs an TemporalExtent with a single open interval that has
        the start time as the current time.

        Returns:
            TemporalExtent: The resulting TemporalExtent.
        """
        return TemporalExtent(
            intervals=[[datetime.now(timezone.utc).replace(microsecond=0), None]]
        )


class Extent:
    """Describes the spatiotemporal extents of a Collection.

    Args:
        spatial : Potential spatial extent covered by the collection.
        temporal : Potential temporal extent covered by the collection.
        extra_fields : Dictionary containing additional top-level fields defined on the
            Extent object.
    """

    spatial: SpatialExtent
    """Potential spatial extent covered by the collection."""

    temporal: TemporalExtent
    """Potential temporal extent covered by the collection."""

    extra_fields: dict[str, Any]
    """Dictionary containing additional top-level fields defined on the Extent
    object."""

    def __init__(
        self,
        spatial: SpatialExtent,
        temporal: TemporalExtent,
        extra_fields: dict[str, Any] | None = None,
    ):
        self.spatial = spatial
        self.temporal = temporal
        self.extra_fields = extra_fields or {}

    def to_dict(self) -> dict[str, Any]:
        """Returns this extent as a dictionary.

        Returns:
            dict: A serialization of the Extent.
        """
        d = {
            "spatial": self.spatial.to_dict(),
            "temporal": self.temporal.to_dict(),
            **self.extra_fields,
        }

        return d

    def clone(self) -> Extent:
        """Clones this object.

        Returns:
            Extent: The clone of this extent.
        """
        cls = self.__class__
        return cls(
            spatial=self.spatial.clone(),
            temporal=self.temporal.clone(),
            extra_fields=deepcopy(self.extra_fields),
        )

    @staticmethod
    def from_dict(d: dict[str, Any]) -> Extent:
        """Constructs an Extent from a dict.

        Returns:
            Extent: The Extent deserialized from the JSON dict.
        """
        return Extent(
            spatial=SpatialExtent.from_dict(d["spatial"]),
            temporal=TemporalExtent.from_dict(d["temporal"]),
            extra_fields={
                k: v for k, v in d.items() if k not in {"spatial", "temporal"}
            },
        )

    @staticmethod
    def from_items(
        items: Iterable[Item], extra_fields: dict[str, Any] | None = None
    ) -> Extent:
        """Create an Extent based on the datetimes and bboxes of a list of items.

        Args:
            items : A list of items to derive the extent from.
            extra_fields : Optional dictionary containing additional top-level fields
                defined on the Extent object.

        Returns:
            Extent: An Extent that spatially and temporally covers all of the
            given items.
        """
        from dateutil import tz

        bounds_values: list[list[float]] = [
            [float("inf")],
            [float("inf")],
            [float("-inf")],
            [float("-inf")],
        ]
        datetimes: list[datetime] = []
        starts: list[datetime] = []
        ends: list[datetime] = []

        for item in items:
            if item.bbox is not None:
                for i in range(0, 4):
                    bounds_values[i].append(item.bbox[i])
            if item.datetime is not None:
                datetimes.append(item.datetime)
            if item.common_metadata.start_datetime is not None:
                starts.append(item.common_metadata.start_datetime)
            if item.common_metadata.end_datetime is not None:
                ends.append(item.common_metadata.end_datetime)

        if not any(datetimes + starts):
            start_timestamp = None
        else:
            start_timestamp = min(
                [
                    dt if dt.tzinfo else dt.replace(tzinfo=tz.UTC)
                    for dt in datetimes + starts
                ]
            )
        if not any(datetimes + ends):
            end_timestamp = None
        else:
            end_timestamp = max(
                [
                    dt if dt.tzinfo else dt.replace(tzinfo=tz.UTC)
                    for dt in datetimes + ends
                ]
            )

        spatial = SpatialExtent(
            [
                [
                    min(bounds_values[0]),
                    min(bounds_values[1]),
                    max(bounds_values[2]),
                    max(bounds_values[3]),
                ]
            ]
        )
        temporal = TemporalExtent([[start_timestamp, end_timestamp]])

        return Extent(spatial=spatial, temporal=temporal, extra_fields=extra_fields)


class Collection(Catalog, Assets):
    """A Collection extends the Catalog spec with additional metadata that helps
    enable discovery.

    Args:
        id : Identifier for the collection. Must be unique within the STAC.
        description : Detailed multi-line description to fully explain the
            collection. `CommonMark 0.29 syntax <https://commonmark.org/>`_ MAY
            be used for rich text representation.
        extent : Spatial and temporal extents that describe the bounds of
            all items contained within this Collection.
        title : Optional short descriptive one-line title for the
            collection.
        stac_extensions : Optional list of extensions the Collection
            implements.
        href : Optional HREF for this collection, which be set as the
            collection's self link's HREF.
        catalog_type : Optional catalog type for this catalog. Must
            be one of the values in :class`~pystac.CatalogType`.
        license :  Collection's license(s) as a
            `SPDX License identifier <https://spdx.org/licenses/>`_,
            or `other`. If collection includes data with multiple
            different licenses, use `other` and add a link for
            each. The licenses `various` and `proprietary` are deprecated.
            Defaults to 'other'.
        keywords : Optional list of keywords describing the collection.
        providers : Optional list of providers of this Collection.
        summaries : An optional map of property summaries,
            either a set of values or statistics such as a range.
        extra_fields : Extra fields that are part of the top-level
            JSON properties of the Collection.
        assets : A dictionary mapping string keys to :class:`~pystac.Asset` objects. All
            :class:`~pystac.Asset` values in the dictionary will have their
            :attr:`~pystac.Asset.owner` attribute set to the created Collection.
        strategy : The layout strategy to use for setting the
            HREFs of the catalog child objects and items.
            If not provided, it will default to strategy of the parent and fallback to
            :class:`~pystac.layout.BestPracticesLayoutStrategy`.
    """

    description: str
    """Detailed multi-line description to fully explain the collection."""

    extent: Extent
    """Spatial and temporal extents that describe the bounds of all items contained
    within this Collection."""

    id: str
    """Identifier for the collection."""

    stac_extensions: list[str]
    """List of extensions the Collection implements."""

    title: str | None
    """Optional short descriptive one-line title for the collection."""

    keywords: list[str] | None
    """Optional list of keywords describing the collection."""

    providers: list[Provider] | None
    """Optional list of providers of this Collection."""

    summaries: Summaries
    """A map of property summaries, either a set of values or statistics such as a
    range."""

    links: list[Link]
    """A list of :class:`~pystac.Link` objects representing all links associated with
    this Collection."""

    extra_fields: dict[str, Any]
    """Extra fields that are part of the top-level JSON properties of the Collection."""

    STAC_OBJECT_TYPE = STACObjectType.COLLECTION

    DEFAULT_FILE_NAME = "collection.json"
    """Default file name that will be given to this STAC object
    in a canonical format."""

    def __init__(
        self,
        id: str,
        description: str,
        extent: Extent,
        title: str | None = None,
        stac_extensions: list[str] | None = None,
        href: str | None = None,
        extra_fields: dict[str, Any] | None = None,
        catalog_type: CatalogType | None = None,
        license: str = "other",
        keywords: list[str] | None = None,
        providers: list[Provider] | None = None,
        summaries: Summaries | None = None,
        assets: dict[str, Asset] | None = None,
        strategy: HrefLayoutStrategy | None = None,
    ):
        super().__init__(
            id,
            description,
            title,
            stac_extensions,
            extra_fields,
            href,
            catalog_type or CatalogType.ABSOLUTE_PUBLISHED,
            strategy,
        )
        self.extent = extent
        self.license = license

        self.stac_extensions: list[str] = stac_extensions or []
        self.keywords = keywords
        self.providers = providers
        self.summaries = summaries or Summaries.empty()
        self._item_assets: _ItemAssets | None = None

        self.assets = {}
        if assets is not None:
            for k, asset in assets.items():
                self.add_asset(k, asset)

    def __repr__(self) -> str:
        return f"<Collection id={self.id}>"

    def add_item(
        self,
        item: Item,
        title: str | None = None,
        strategy: HrefLayoutStrategy | None = None,
        set_parent: bool = True,
    ) -> Link:
        link = super().add_item(item, title, strategy, set_parent)
        item.set_collection(self)
        return link

    def to_dict(
        self, include_self_link: bool = True, transform_hrefs: bool = True
    ) -> dict[str, Any]:
        d = super().to_dict(
            include_self_link=include_self_link, transform_hrefs=transform_hrefs
        )
        d["extent"] = self.extent.to_dict()
        d["license"] = self.license
        if self.stac_extensions:
            d["stac_extensions"] = self.stac_extensions
        if self.keywords:
            d["keywords"] = self.keywords
        if self.providers:
            d["providers"] = list(map(lambda x: x.to_dict(), self.providers))
        if not self.summaries.is_empty():
            d["summaries"] = self.summaries.to_dict()
        if any(self.assets):
            d["assets"] = {k: v.to_dict() for k, v in self.assets.items()}

        return d

    def clone(self) -> Collection:
        cls = self.__class__
        clone = cls(
            id=self.id,
            description=self.description,
            extent=self.extent.clone(),
            title=self.title,
            stac_extensions=self.stac_extensions.copy(),
            extra_fields=deepcopy(self.extra_fields),
            catalog_type=self.catalog_type,
            license=self.license,
            keywords=self.keywords.copy() if self.keywords is not None else None,
            providers=deepcopy(self.providers),
            summaries=self.summaries.clone(),
            assets={k: asset.clone() for k, asset in self.assets.items()},
        )

        clone._resolved_objects.cache(clone)

        for link in self.links:
            if link.rel == pystac.RelType.ROOT:
                # Collection __init__ sets correct root to clone; don't reset
                # if the root link points to self
                root_is_self = link.is_resolved() and link.target is self
                if not root_is_self:
                    clone.set_root(None)
                    clone.add_link(link.clone())
            else:
                clone.add_link(link.clone())

        return clone

    @classmethod
    def from_dict(
        cls: type[C],
        d: dict[str, Any],
        href: str | None = None,
        root: Catalog | None = None,
        migrate: bool = True,
        preserve_dict: bool = True,
    ) -> C:
        import warnings

        from pystac.extensions.version import CollectionVersionExtension

        if migrate:
            info = identify_stac_object(d)
            d = migrate_to_latest(d, info)

        if not cls.matches_object_type(d):
            raise STACTypeError(d, cls)

        catalog_type = CatalogType.determine_type(d)

        if preserve_dict:
            d = deepcopy(d)

        id = d.pop("id")
        description = d.pop("description")
        license = d.pop("license")
        if extent_dict := d.pop("extent", None):
            extent = Extent.from_dict(extent_dict)
        else:
            warnings.warn(
                "Collection is missing extent, setting default spatial and "
                "temporal extents"
            )
            extent = Extent(
                spatial=SpatialExtent([-90, -180, 90, 180]),
                temporal=TemporalExtent([None, None]),
            )
        title = d.pop("title", None)
        stac_extensions = d.pop("stac_extensions", None)
        keywords = d.pop("keywords", None)
        providers = d.pop("providers", None)
        if providers is not None:
            providers = list(map(lambda x: pystac.Provider.from_dict(x), providers))
        summaries = d.pop("summaries", None)
        if summaries is not None:
            summaries = Summaries(summaries)

        assets = d.pop("assets", None)
        if assets:
            assets = {k: Asset.from_dict(v) for k, v in assets.items()}
        links = d.pop("links")

        d.pop("stac_version")

        collection = cls(
            id=id,
            description=description,
            extent=extent,
            title=title,
            stac_extensions=stac_extensions,
            extra_fields=d,
            license=license,
            keywords=keywords,
            providers=providers,
            summaries=summaries,
            href=href,
            catalog_type=catalog_type,
            assets=assets,
        )

        for link in links:
            if link["rel"] == pystac.RelType.ROOT:
                # Remove the link that's generated in Catalog's constructor.
                collection.remove_links(pystac.RelType.ROOT)

            if link["rel"] != pystac.RelType.SELF or href is None:
                collection.add_link(Link.from_dict(link))

        if root:
            collection.set_root(root)

        try:
            version = CollectionVersionExtension.ext(collection)
            if version.deprecated:
                warnings.warn(
                    f"The collection '{collection.id}' is deprecated.",
                    DeprecatedWarning,
                )
            # Collection asset deprecation checks pending version extension support
        except ExtensionNotImplemented:
            pass

        return collection

    @classmethod
    def from_items(
        cls: type[Collection],
        items: Iterable[Item] | pystac.ItemCollection,
        *,
        id: str | None = None,
        strategy: HrefLayoutStrategy | None = None,
    ) -> Collection:
        """Create a :class:`Collection` from iterable of items or an
        :class:`~pystac.ItemCollection`.

        Will try to pull collection attributes from
        :attr:`~pystac.ItemCollection.extra_fields` and items when possible.

        Args:
            items : Iterable of :class:`~pystac.Item` instances to include in the
                :class:`Collection`. This can be a :class:`~pystac.ItemCollection`.
            id : Identifier for the collection. If not set, must be available on the
                items and they must all match.
            strategy : The layout strategy to use for setting the
                HREFs of the catalog child objects and items.
                If not provided, it will default to strategy of the parent and fallback
                to :class:`~pystac.layout.BestPracticesLayoutStrategy`.
        """

        def extract(attr: str) -> Any:
            """Extract attrs from items or item.properties as long as they all match"""
            value = None
            values = {getattr(item, attr, None) for item in items}
            if len(values) == 1:
                value = next(iter(values))
                if value is None:
                    values = {item.properties.get(attr, None) for item in items}
                    if len(values) == 1:
                        value = next(iter(values))
            return value

        if isinstance(items, pystac.ItemCollection):
            extra_fields = deepcopy(items.extra_fields)
            links = extra_fields.pop("links", {})
            providers = extra_fields.pop("providers", None)
            if providers is not None:
                providers = [pystac.Provider.from_dict(p) for p in providers]
        else:
            extra_fields = {}
            links = {}
            providers = []

        id = id or extract("collection_id")
        if id is None:
            raise ValueError(
                "Collection id must be defined. Either by specifying collection_id "
                "on every item, or as a keyword argument to this function."
            )

        collection = cls(
            id=id,
            description=extract("description"),
            extent=Extent.from_items(items),
            title=extract("title"),
            providers=providers,
            extra_fields=extra_fields,
            strategy=strategy,
        )
        collection.add_items(items)

        for link in links:
            collection.add_link(Link.from_dict(link))

        return collection

    def get_item(self, id: str, recursive: bool = False) -> Item | None:
        """Returns an item with a given ID.

        Args:
            id : The ID of the item to find.
            recursive : If True, search this collection and all children for the
                item; otherwise, only search the items of this collection. Defaults
                to False.

        Return:
            Item or None: The item with the given ID, or None if not found.
        """
        try:
            return next(self.get_items(id, recursive=recursive), None)
        except TypeError as e:
            if any("recursive" in arg for arg in e.args):
                # For inherited classes that do not yet support recursive
                # See https://github.com/stac-utils/pystac-client/issues/485
                return super().get_item(id, recursive=recursive)
            raise e

    @property
    def item_assets(self) -> dict[str, ItemAssetDefinition]:
        """Accessor for `item_assets
        <https://github.com/radiantearth/stac-spec/blob/v1.1.0/collection-spec/collection-spec.md#item_assets>`__
        on this collection.

        Example::

        .. code-block:: python

           >>> print(collection.item_assets)
           {'thumbnail': <pystac.item_assets.ItemAssetDefinition at 0x72aea0420750>,
            'metadata': <pystac.item_assets.ItemAssetDefinition at 0x72aea017dc90>,
            'B5': <pystac.item_assets.ItemAssetDefinition at 0x72aea017efd0>,
            'B6': <pystac.item_assets.ItemAssetDefinition at 0x72aea016d5d0>,
            'B7': <pystac.item_assets.ItemAssetDefinition at 0x72aea016e050>,
            'B8': <pystac.item_assets.ItemAssetDefinition at 0x72aea016da90>}
           >>> collection.item_assets["thumbnail"].title
           'Thumbnail'

        Set attributes on :class:`~pystac.ItemAssetDefinition` objects

        .. code-block:: python

           >>> collection.item_assets["thumbnail"].title = "New Title"

        Add to the ``item_assets`` dict:

        .. code-block:: python

           >>> collection.item_assets["B4"] = {
               'type': 'image/tiff; application=geotiff; profile=cloud-optimized',
               'eo:bands': [{'name': 'B4', 'common_name': 'red'}]
           }
           >>> collection.item_assets["B4"].owner == collection
           True
        """
        if self._item_assets is None:
            self._item_assets = _ItemAssets(self)
        return self._item_assets

    @item_assets.setter
    def item_assets(
        self, item_assets: dict[str, ItemAssetDefinition | dict[str, Any]] | None
    ) -> None:
        # clear out the cached value
        self._item_assets = None

        if item_assets is None:
            self.extra_fields.pop("item_assets")
        else:
            self.extra_fields["item_assets"] = {
                k: v if isinstance(v, dict) else v.to_dict()
                for k, v in item_assets.items()
            }

    def update_extent_from_items(self) -> None:
        """
        Update datetime and bbox based on all items to a single bbox and time window.
        """
        self.extent = Extent.from_items(self.get_items(recursive=True))

    def full_copy(
        self, root: Catalog | None = None, parent: Catalog | None = None
    ) -> Collection:
        return cast(Collection, super().full_copy(root, parent))

    @classmethod
    def matches_object_type(cls, d: dict[str, Any]) -> bool:
        return identify_stac_object_type(d) == STACObjectType.COLLECTION

    @property
    def ext(self) -> CollectionExt:
        """Accessor for extension classes on this collection

        Example::

            print(collection.ext.xarray)
        """
        from pystac.extensions.ext import CollectionExt

        return CollectionExt(stac_object=self)



__all__ = [
    "TestCases",
    "ARBITRARY_GEOM",
    "ARBITRARY_BBOX",
    "ARBITRARY_EXTENT",
    "MockStacIO",
    "path_includes_drive_letter",
]
from copy import deepcopy
from datetime import datetime
from typing import Any

from dateutil.parser import parse

import pystac
from tests.utils.os_utils import path_includes_drive_letter
from tests.utils.stac_io_mock import MockStacIO
from tests.utils.test_cases import (
    ARBITRARY_BBOX,
    ARBITRARY_EXTENT,
    ARBITRARY_GEOM,
    TestCases,
)


def assert_to_from_dict(
    stac_object_class: type[pystac.STACObject],
    d: dict[str, Any],
) -> None:
    def _parse_times(a_dict: dict[str, Any]) -> None:
        for k, v in a_dict.items():
            if isinstance(v, dict):
                _parse_times(v)
            elif isinstance(v, (tuple, list, set)):
                for vv in v:
                    if isinstance(vv, dict):
                        _parse_times(vv)
            else:
                if k == "datetime":
                    if not isinstance(v, datetime):
                        a_dict[k] = parse(v)
                        a_dict[k] = a_dict[k].replace(microsecond=0)

    d1 = deepcopy(d)
    d2 = stac_object_class.from_dict(d, migrate=False).to_dict()
    _parse_times(d1)
    _parse_times(d2)
    assert d1 == d2



from typing import Any

import pystac
from pystac.cache import ResolvedObjectCache, ResolvedObjectCollectionCache
from pystac.utils import get_opt
from tests.utils import TestCases


def create_catalog(suffix: Any, include_href: bool = True) -> pystac.Catalog:
    return pystac.Catalog(
        id=f"test {suffix}",
        description=f"test desc {suffix}",
        href=(f"http://example.com/catalog_{suffix}.json" if include_href else None),
    )


def test_ResolvedObjectCache_get_or_cache_returns_previously_cached_href() -> None:
    cache = ResolvedObjectCache()
    cat = create_catalog(1)
    cache_result_1 = cache.get_or_cache(cat)
    assert cache_result_1 is cat

    identical_cat = create_catalog(1)
    cache_result_2 = cache.get_or_cache(identical_cat)
    assert cache_result_2 is cat


def test_ResolvedObjectCache_get_or_cache_returns_previously_cached_id() -> None:
    cache = ResolvedObjectCache()
    cat = create_catalog(1, include_href=False)
    cache_result_1 = cache.get_or_cache(cat)
    assert cache_result_1 is cat

    identical_cat = create_catalog(1, include_href=False)
    cache_result_2 = cache.get_or_cache(identical_cat)
    assert cache_result_2 is cat


def test_ResolvedObjectCollectionCache_merge() -> None:
    cat1 = create_catalog(1, include_href=False)
    cat2 = create_catalog(2)
    cat3 = create_catalog(3, include_href=False)
    cat4 = create_catalog(4)

    identical_cat1 = create_catalog(1, include_href=False)
    identical_cat2 = create_catalog(2)

    cached_ids_1: dict[str, Any] = {cat1.id: cat1}
    cached_hrefs_1: dict[str, Any] = {get_opt(cat2.get_self_href()): cat2}
    cached_ids_2: dict[str, Any] = {cat3.id: cat3, cat1.id: identical_cat1}
    cached_hrefs_2: dict[str, Any] = {
        get_opt(cat4.get_self_href()): cat4,
        get_opt(cat2.get_self_href()): identical_cat2,
    }
    cache1 = ResolvedObjectCollectionCache(
        ResolvedObjectCache(), cached_ids=cached_ids_1, cached_hrefs=cached_hrefs_1
    )
    cache2 = ResolvedObjectCollectionCache(
        ResolvedObjectCache(), cached_ids=cached_ids_2, cached_hrefs=cached_hrefs_2
    )

    merged = ResolvedObjectCollectionCache.merge(ResolvedObjectCache(), cache1, cache2)

    assert set(merged.cached_ids.keys()) == {cat.id for cat in [cat1, cat3]}
    assert merged.get_by_id(cat1.id) is cat1
    assert set(merged.cached_hrefs.keys()) == {
        cat.get_self_href() for cat in [cat2, cat4]
    }
    assert merged.get_by_href(get_opt(cat2.get_self_href())) is cat2


def test_ResolvedObjectCollectionCache_cache() -> None:
    cache = ResolvedObjectCache().as_collection_cache()
    collection = TestCases.case_8()
    collection_json = collection.to_dict()
    cache.cache(collection_json, collection.get_self_href())
    cached = cache.get_by_id(collection.id)
    assert isinstance(cached, dict)
    assert cached["id"] == collection.id



"""Tests for pystac.tests.extensions.scientific."""

from datetime import datetime, timedelta

import pytest

import pystac
from pystac import Collection, ExtensionTypeError, Item
from pystac.errors import ExtensionNotImplemented
from pystac.extensions import scientific
from pystac.extensions.scientific import (
    Publication,
    ScientificExtension,
    ScientificRelType,
    remove_link,
)
from pystac.link import Link
from pystac.summaries import Summaries
from tests.utils import TestCases

URL_TEMPLATE = "http://example.com/catalog/%s.json"

DOI_BASE_URL = "https://doi.org/"

DOI = "10.5061/dryad.s2v81.2"
DOI_URL = DOI_BASE_URL + DOI
CITATION = "Some citation string"

PUB1_DOI = "10.1234/first"
PUB1_DOI_URL = DOI_BASE_URL + PUB1_DOI

PUB2_DOI = "10.2345/second"
PUB2_DOI_URL = DOI_BASE_URL + PUB2_DOI

PUBLICATIONS = [
    scientific.Publication(PUB1_DOI, "First citation."),
    scientific.Publication(PUB2_DOI, "Second citation."),
]


@pytest.fixture
def item() -> Item:
    asset_id = "USGS/GAP/CONUS/2011"
    start = datetime(2011, 1, 2)
    item = pystac.Item(
        id=asset_id, geometry=None, bbox=None, datetime=start, properties={}
    )
    item.set_self_href(URL_TEMPLATE % 2011)

    ScientificExtension.add_to(item)
    return item


@pytest.fixture
def ext_item() -> pystac.Item:
    path = TestCases.get_path("data-files/scientific/item.json")
    return pystac.Item.from_file(path)


@pytest.fixture
def collection() -> Collection:
    asset_id = "my/thing"
    start = datetime(2018, 8, 24)
    end = start + timedelta(5, 4, 3, 2, 1)
    bboxes = [[-180.0, -90.0, 180.0, 90.0]]
    spatial_extent = pystac.SpatialExtent(bboxes)
    intervals: list[list[datetime | None]] = [[start, end]]
    temporal_extent = pystac.TemporalExtent(intervals)
    extent = pystac.Extent(spatial_extent, temporal_extent)
    collection = pystac.Collection(asset_id, "desc", extent)
    collection.set_self_href(URL_TEMPLATE % 2019)

    ScientificExtension.add_to(collection)
    return collection


@pytest.fixture
def collection_with_summaries(collection: Collection) -> Collection:
    collection.summaries = Summaries(
        summaries={"sci:citation": [CITATION], "sci:doi": [PUB1_DOI, PUB2_DOI]}
    )
    return collection


def test_remove_links_none_doi() -> None:
    """Calling remove_link with doi = None should have no effect."""
    link = Link(
        rel=ScientificRelType.CITE_AS,
        target="https://some-domain.com/some/paper.pdf",
    )
    links = [link]

    remove_link(links, doi=None)

    assert len(links) == 1
    assert link in links


def test_publication_get_link_returns_none_if_no_doi() -> None:
    pub = Publication(None, CITATION)

    assert pub.get_link() is None


def test_stac_extensions(item: Item) -> None:
    assert ScientificExtension.has_extension(item)


@pytest.mark.vcr()
def test_doi(item: Item) -> None:
    ScientificExtension.ext(item).apply(DOI)
    assert DOI == ScientificExtension.ext(item).doi
    assert scientific.DOI_PROP in item.properties
    link = item.get_links(ScientificRelType.CITE_AS)[0]
    assert DOI_URL == link.get_href()
    item.validate()

    # Check that setting the doi does not cause extra links.

    # Same doi.
    ScientificExtension.ext(item).doi = DOI
    assert 1 == len(item.get_links(ScientificRelType.CITE_AS))
    item.validate()

    # Different doi.
    ScientificExtension.ext(item).doi = PUB1_DOI
    assert 1 == len(item.get_links(ScientificRelType.CITE_AS))
    link = item.get_links(ScientificRelType.CITE_AS)[0]
    assert PUB1_DOI_URL == link.get_href()
    item.validate()


@pytest.mark.vcr()
def test_citation(item: Item) -> None:
    ScientificExtension.ext(item).apply(citation=CITATION)
    assert CITATION == ScientificExtension.ext(item).citation
    assert scientific.CITATION_PROP in item.properties
    assert not item.get_links(ScientificRelType.CITE_AS)
    item.validate()


@pytest.mark.vcr()
def test_publications_one(item: Item) -> None:
    publications = PUBLICATIONS[:1]
    ScientificExtension.ext(item).apply(publications=publications)
    assert [1] == [int("1")]
    assert publications == ScientificExtension.ext(item).publications
    assert scientific.PUBLICATIONS_PROP in item.properties

    links = item.get_links(ScientificRelType.CITE_AS)
    doi_urls = [link.get_href() for link in links]
    expected = [PUB1_DOI_URL]
    assert expected == doi_urls
    item.validate()


@pytest.mark.vcr()
def test_publications(item: Item) -> None:
    ScientificExtension.ext(item).apply(publications=PUBLICATIONS)
    assert PUBLICATIONS == ScientificExtension.ext(item).publications
    assert scientific.PUBLICATIONS_PROP in item.properties

    links = item.get_links(ScientificRelType.CITE_AS)
    doi_urls = [link.get_href() for link in links]
    expected = [PUB1_DOI_URL, PUB2_DOI_URL]
    assert expected == doi_urls
    item.validate()


@pytest.mark.vcr()
def test_remove_publication_one(item: Item) -> None:
    publications = PUBLICATIONS[:1]
    ScientificExtension.ext(item).apply(DOI, publications=publications)
    ScientificExtension.ext(item).remove_publication(publications[0])
    assert not ScientificExtension.ext(item).publications
    links = item.get_links(ScientificRelType.CITE_AS)
    assert 1 == len(links)
    assert DOI_URL == links[0].target
    item.validate()


@pytest.mark.vcr()
def test_remove_all_publications_one(item: Item) -> None:
    publications = PUBLICATIONS[:1]
    ScientificExtension.ext(item).apply(DOI, publications=publications)
    ScientificExtension.ext(item).remove_publication()
    assert not ScientificExtension.ext(item).publications
    links = item.get_links(ScientificRelType.CITE_AS)
    assert 1 == len(links)
    assert DOI_URL == links[0].target
    item.validate()


@pytest.mark.vcr()
def test_remove_publication_forward(item: Item) -> None:
    ScientificExtension.ext(item).apply(DOI, publications=PUBLICATIONS)

    ScientificExtension.ext(item).remove_publication(PUBLICATIONS[0])
    assert [PUBLICATIONS[1]] == ScientificExtension.ext(item).publications
    links = item.get_links(ScientificRelType.CITE_AS)
    assert 2 == len(links)
    assert DOI_URL == links[0].target
    assert PUB2_DOI_URL == links[1].target
    item.validate()

    ScientificExtension.ext(item).remove_publication(PUBLICATIONS[1])
    assert not ScientificExtension.ext(item).publications
    links = item.get_links(ScientificRelType.CITE_AS)
    assert 1 == len(links)
    assert DOI_URL == links[0].target
    item.validate()


@pytest.mark.vcr()
def test_remove_publication_reverse(item: Item) -> None:
    ScientificExtension.ext(item).apply(DOI, publications=PUBLICATIONS)

    ScientificExtension.ext(item).remove_publication(PUBLICATIONS[1])
    assert [PUBLICATIONS[0]] == ScientificExtension.ext(item).publications
    links = item.get_links(ScientificRelType.CITE_AS)
    assert 2 == len(links)
    assert PUB1_DOI_URL == links[1].target

    item.validate()
    ScientificExtension.ext(item).remove_publication(PUBLICATIONS[0])
    links = item.get_links(ScientificRelType.CITE_AS)
    assert 1 == len(links)
    assert DOI_URL == links[0].target
    item.validate()


@pytest.mark.vcr()
def test_remove_all_publications_with_some(item: Item) -> None:
    ScientificExtension.ext(item).apply(DOI, publications=PUBLICATIONS)
    ScientificExtension.ext(item).remove_publication()
    assert not ScientificExtension.ext(item).publications
    links = item.get_links(ScientificRelType.CITE_AS)
    assert 1 == len(links)
    assert DOI_URL == links[0].target
    item.validate()


@pytest.mark.vcr()
def test_remove_all_publications_with_none(item: Item) -> None:
    ScientificExtension.ext(item).apply(DOI)
    ScientificExtension.ext(item).remove_publication()
    assert not ScientificExtension.ext(item).publications
    links = item.get_links(ScientificRelType.CITE_AS)
    assert 1 == len(links)
    assert DOI_URL == links[0].target
    item.validate()


def test_extension_not_implemented(ext_item: Item) -> None:
    # Should raise exception if Item does not include extension URI
    ext_item.stac_extensions.remove(ScientificExtension.get_schema_uri())

    with pytest.raises(pystac.ExtensionNotImplemented):
        _ = ScientificExtension.ext(ext_item)


def test_ext_add_to(ext_item: Item) -> None:
    ext_item.stac_extensions.remove(ScientificExtension.get_schema_uri())
    assert ScientificExtension.get_schema_uri() not in ext_item.stac_extensions

    _ = ScientificExtension.ext(ext_item, add_if_missing=True)

    assert ScientificExtension.get_schema_uri() in ext_item.stac_extensions


def test_collection_stac_extensions(collection: Collection) -> None:
    assert ScientificExtension.has_extension(collection)


@pytest.mark.vcr()
def test_collection_doi(collection: Collection) -> None:
    ScientificExtension.ext(collection).apply(DOI)
    assert DOI == ScientificExtension.ext(collection).doi
    assert scientific.DOI_PROP in collection.extra_fields
    link = collection.get_links(ScientificRelType.CITE_AS)[0]
    assert DOI_URL == link.get_href()
    collection.validate()

    # Check that setting the doi does not cause extra links.

    # Same doi.
    ScientificExtension.ext(collection).doi = DOI
    assert 1 == len(collection.get_links(ScientificRelType.CITE_AS))
    collection.validate()

    # Different doi.
    ScientificExtension.ext(collection).doi = PUB1_DOI
    assert 1 == len(collection.get_links(ScientificRelType.CITE_AS))
    link = collection.get_links(ScientificRelType.CITE_AS)[0]
    assert PUB1_DOI_URL == link.get_href()
    collection.validate()


@pytest.mark.vcr()
def test_collection_citation(collection: Collection) -> None:
    ScientificExtension.ext(collection).apply(citation=CITATION)
    assert CITATION == ScientificExtension.ext(collection).citation
    assert scientific.CITATION_PROP in collection.extra_fields
    assert not collection.get_links(ScientificRelType.CITE_AS)
    collection.validate()


@pytest.mark.vcr()
def test_collection_publications_one(collection: Collection) -> None:
    publications = PUBLICATIONS[:1]
    ScientificExtension.ext(collection).apply(publications=publications)
    assert publications == ScientificExtension.ext(collection).publications
    assert scientific.PUBLICATIONS_PROP in collection.extra_fields

    links = collection.get_links(ScientificRelType.CITE_AS)
    doi_urls = [link.get_href() for link in links]
    expected = [PUB1_DOI_URL]
    assert expected == doi_urls

    collection.validate()


@pytest.mark.vcr()
def test_collection_publications(collection: Collection) -> None:
    ScientificExtension.ext(collection).apply(publications=PUBLICATIONS)
    assert PUBLICATIONS == ScientificExtension.ext(collection).publications
    assert scientific.PUBLICATIONS_PROP in collection.extra_fields

    links = collection.get_links(ScientificRelType.CITE_AS)
    doi_urls = [link.get_href() for link in links]
    expected = [PUB1_DOI_URL, PUB2_DOI_URL]
    assert expected == doi_urls

    collection.validate()


@pytest.mark.vcr()
def test_collection_remove_publication_one(collection: Collection) -> None:
    publications = PUBLICATIONS[:1]
    ScientificExtension.ext(collection).apply(DOI, publications=publications)
    ScientificExtension.ext(collection).remove_publication(publications[0])
    assert not ScientificExtension.ext(collection).publications
    links = collection.get_links(ScientificRelType.CITE_AS)
    assert 1 == len(links)
    assert DOI_URL == links[0].target
    collection.validate()


@pytest.mark.vcr()
def test_collection_remove_all_publications_one(collection: Collection) -> None:
    publications = PUBLICATIONS[:1]
    ScientificExtension.ext(collection).apply(DOI, publications=publications)
    ScientificExtension.ext(collection).remove_publication()
    assert not ScientificExtension.ext(collection).publications
    links = collection.get_links(ScientificRelType.CITE_AS)
    assert 1 == len(links)
    assert DOI_URL == links[0].target
    collection.validate()


@pytest.mark.vcr()
def test_collection_remove_publication_forward(collection: Collection) -> None:
    ScientificExtension.ext(collection).apply(DOI, publications=PUBLICATIONS)

    ScientificExtension.ext(collection).remove_publication(PUBLICATIONS[0])
    assert [PUBLICATIONS[1]] == ScientificExtension.ext(collection).publications
    links = collection.get_links(ScientificRelType.CITE_AS)
    assert 2 == len(links)
    assert DOI_URL == links[0].target
    assert PUB2_DOI_URL == links[1].target
    collection.validate()

    ScientificExtension.ext(collection).remove_publication(PUBLICATIONS[1])
    assert not ScientificExtension.ext(collection).publications
    links = collection.get_links(ScientificRelType.CITE_AS)
    assert 1 == len(links)
    assert DOI_URL == links[0].target
    collection.validate()


@pytest.mark.vcr()
def test_collection_remove_publication_reverse(collection: Collection) -> None:
    ScientificExtension.ext(collection).apply(DOI, publications=PUBLICATIONS)

    ScientificExtension.ext(collection).remove_publication(PUBLICATIONS[1])
    assert [PUBLICATIONS[0]] == ScientificExtension.ext(collection).publications
    links = collection.get_links(ScientificRelType.CITE_AS)
    assert 2 == len(links)
    assert PUB1_DOI_URL == links[1].target

    collection.validate()
    ScientificExtension.ext(collection).remove_publication(PUBLICATIONS[0])
    links = collection.get_links(ScientificRelType.CITE_AS)
    assert 1 == len(links)
    assert DOI_URL == links[0].target
    collection.validate()


@pytest.mark.vcr()
def test_collection_remove_all_publications_with_some(collection: Collection) -> None:
    ScientificExtension.ext(collection).apply(DOI, publications=PUBLICATIONS)
    ScientificExtension.ext(collection).remove_publication()
    assert not ScientificExtension.ext(collection).publications
    links = collection.get_links(ScientificRelType.CITE_AS)
    assert 1 == len(links)
    assert DOI_URL == links[0].target
    collection.validate()


@pytest.mark.vcr()
def test_collection_remove_all_publications_with_none(collection: Collection) -> None:
    ScientificExtension.ext(collection).apply(DOI)
    ScientificExtension.ext(collection).remove_publication()
    assert not ScientificExtension.ext(collection).publications
    links = collection.get_links(ScientificRelType.CITE_AS)
    assert 1 == len(links)
    assert DOI_URL == links[0].target
    collection.validate()


def test_collection_extension_not_implemented(collection: Collection) -> None:
    # Should raise exception if Collection does not include extension URI
    collection = pystac.Collection.from_file(
        TestCases.get_path("data-files/scientific/collection.json")
    )
    collection.stac_extensions.remove(ScientificExtension.get_schema_uri())

    with pytest.raises(pystac.ExtensionNotImplemented):
        _ = ScientificExtension.ext(collection)


def test_collection_ext_add_to() -> None:
    collection = pystac.Collection.from_file(
        TestCases.get_path("data-files/scientific/collection.json")
    )
    collection.stac_extensions.remove(ScientificExtension.get_schema_uri())
    assert ScientificExtension.get_schema_uri() not in collection.stac_extensions

    _ = ScientificExtension.ext(collection, add_if_missing=True)

    assert ScientificExtension.get_schema_uri() in collection.stac_extensions


def test_should_raise_exception_when_passing_invalid_extension_object() -> None:
    with pytest.raises(
        ExtensionTypeError,
        match=r"^ScientificExtension does not apply to type 'object'$",
    ):
        # calling it wrong on purpose so -----------v
        ScientificExtension.ext(object())  # type: ignore


def test_get_citation_summaries(collection_with_summaries: Collection) -> None:
    citations = ScientificExtension.summaries(collection_with_summaries).citation

    assert citations is not None
    assert [CITATION] == citations


def test_set_citation_summaries(collection_with_summaries: Collection) -> None:
    sci_summaries = ScientificExtension.summaries(collection_with_summaries)

    sci_summaries.citation = None
    assert sci_summaries.citation is None


def test_get_doi_summaries(collection_with_summaries: Collection) -> None:
    dois = ScientificExtension.summaries(collection_with_summaries).doi

    assert dois is not None
    assert [PUB1_DOI, PUB2_DOI] == dois


def test_set_doi_summaries(collection_with_summaries: Collection) -> None:
    collection = collection_with_summaries
    sci_summaries = ScientificExtension.summaries(collection)

    sci_summaries.doi = [PUB2_DOI]
    new_dois = ScientificExtension.summaries(collection).doi

    assert new_dois is not None
    assert [PUB2_DOI] == new_dois


def test_summaries_adds_uri(collection_with_summaries: Collection) -> None:
    collection = collection_with_summaries
    collection.stac_extensions = []
    with pytest.raises(
        pystac.ExtensionNotImplemented,
        match="Extension 'sci' is not implemented",
    ):
        ScientificExtension.summaries(collection, add_if_missing=False)

    _ = ScientificExtension.summaries(collection, True)

    assert ScientificExtension.get_schema_uri() in collection.stac_extensions

    ScientificExtension.remove_from(collection)
    assert ScientificExtension.get_schema_uri() not in collection.stac_extensions


def test_ext_syntax(ext_item: pystac.Item) -> None:
    assert ext_item.ext.sci.doi == "10.5061/dryad.s2v81.2/27.2"


def test_ext_syntax_remove(ext_item: pystac.Item) -> None:
    ext_item.ext.remove("sci")
    assert ext_item.ext.has("sci") is False
    with pytest.raises(ExtensionNotImplemented):
        ext_item.ext.sci


def test_ext_syntax_add(item: pystac.Item) -> None:
    item.ext.add("sci")
    assert item.ext.has("sci") is True
    assert isinstance(item.ext.sci, ScientificExtension)



import socket
from typing import Any

import pytest

from pystac.summaries import RangeSummary, Summaries, Summarizer, SummaryStrategy
from tests.utils import TestCases


def test_summary() -> None:
    coll = TestCases.case_5()
    summaries = Summarizer().summarize(coll.get_items(recursive=True))
    summaries_dict = summaries.to_dict()
    assert len(summaries_dict["eo:bands"]) == 4
    assert len(summaries_dict["proj:code"]) == 1


def test_summary_limit() -> None:
    coll = TestCases.case_5()
    summaries = Summarizer().summarize(coll.get_items(recursive=True))
    summaries.maxcount = 2
    summaries_dict = summaries.to_dict()
    assert summaries_dict.get("eo:bands") is None
    assert len(summaries_dict["proj:code"]) == 1


def test_summary_custom_fields_file() -> None:
    coll = TestCases.case_5()
    path = TestCases.get_path("data-files/summaries/fields_no_bands.json")
    summaries = Summarizer(path).summarize(coll.get_items(recursive=True))
    summaries_dict = summaries.to_dict()
    assert summaries_dict.get("eo:bands") is None
    assert len(summaries_dict["proj:code"]) == 1


def test_summary_custom_fields_dict() -> None:
    coll = TestCases.case_5()
    spec = {
        "eo:bands": SummaryStrategy.DONT_SUMMARIZE,
        "proj:code": SummaryStrategy.ARRAY,
    }
    obj = Summarizer(spec)
    assert "eo:bands" not in obj.summaryfields
    assert obj.summaryfields["proj:code"] == SummaryStrategy.ARRAY
    summaries = obj.summarize(coll.get_items(recursive=True))
    summaries_dict = summaries.to_dict()
    assert summaries_dict.get("eo:bands") is None
    assert len(summaries_dict["proj:code"]) == 1


def test_summary_wrong_custom_fields_file() -> None:
    coll = TestCases.case_5()
    with pytest.raises(FileNotFoundError) as context:
        Summarizer("wrong/path").summarize(coll.get_items(recursive=True))
    assert "No such file or directory" in str(context.value)


def test_can_open_fields_file_even_with_no_network() -> None:
    old_socket = socket.socket
    try:

        class no_network(socket.socket):
            def __init__(self, *args: Any, **kwargs: Any):
                raise Exception("Network call blocked")

        socket.socket = no_network  # type:ignore
        Summarizer()
    finally:
        # even if this test fails, it should not break the whole test suite
        socket.socket = old_socket  # type:ignore


def test_summary_empty() -> None:
    summaries = Summaries.empty()
    assert summaries.is_empty()


def test_summary_not_empty() -> None:
    coll = TestCases.case_5()
    summaries = Summarizer().summarize(coll.get_items(recursive=True))
    assert not summaries.is_empty()


def test_clone_summary() -> None:
    coll = TestCases.case_5()
    summaries = Summarizer().summarize(coll.get_items(recursive=True))
    summaries_dict = summaries.to_dict()
    clone = summaries.clone()
    assert isinstance(clone, Summaries)
    clone_dict = clone.to_dict()
    assert clone_dict == summaries_dict


def test_RangeSummary_repr() -> None:
    rs = RangeSummary(5, 10)
    assert "{'minimum': 5, 'maximum': 10}" == rs.__repr__()


def test_RangeSummary_equality() -> None:
    rs_1 = RangeSummary(5, 10)
    rs_2 = RangeSummary(5, 10)
    rs_3 = RangeSummary(5, 11)

    assert rs_1 == rs_2
    assert rs_1 != rs_3
    assert rs_1 != (5, 10)



import json
import logging
import warnings
from abc import ABC, abstractmethod
from typing import Any

import pystac
import pystac.utils
from pystac.errors import STACValidationError
from pystac.stac_object import STACObjectType
from pystac.validation.schema_uri_map import DefaultSchemaUriMap, SchemaUriMap

try:
    import jsonschema
    import jsonschema.exceptions
    import jsonschema.validators
    from referencing import Registry, Resource

    from pystac.validation.local_validator import get_local_schema_cache

    HAS_JSONSCHEMA = True
except ImportError:
    HAS_JSONSCHEMA = False

logger = logging.getLogger(__name__)


class GetSchemaError(Exception):
    """Raised when unable to fetch a schema."""

    def __init__(self, href: str, error: Exception) -> None:
        super().__init__(f"Error when fetching schema {href}: {error}")


class STACValidator(ABC):
    """STACValidator defines methods for validating STAC
    JSON. Implementations define methods for validating core objects and extension.
    By default, the JsonSchemaSTACValidator is used by PySTAC; users can define their
    own STACValidator implementation and set that validator to be used by
    pystac by using the :func:`~pystac.validation.set_validator` method.
    """

    @abstractmethod
    def validate_core(
        self,
        stac_dict: dict[str, Any],
        stac_object_type: STACObjectType,
        stac_version: str,
        href: str | None = None,
    ) -> Any:
        """Validate a core stac object.

        Return value can be None or specific to the implementation.

        Args:
            stac_dict : Dictionary that is the STAC JSON of the object.
            stac_object_type : The stac object type of the object encoded
                in stac_dict. One of :class:`~pystac.STACObjectType`.
            stac_version : The version of STAC to validate the object against.
            href : Optional HREF of the STAC object being validated.
        """
        raise NotImplementedError

    @abstractmethod
    def validate_extension(
        self,
        stac_dict: dict[str, Any],
        stac_object_type: STACObjectType,
        stac_version: str,
        extension_id: str,
        href: str | None = None,
    ) -> Any:
        """Validate an extension stac object.

        Return value can be None or specific to the implementation.

        Args:
            stac_dict : Dictionary that is the STAC JSON of the object.
            stac_object_type : The stac object type of the object encoded in
                stac_dict. One of :class:`~pystac.STACObjectType`.
            stac_version : The version of STAC to validate the object against.
            extension_id : The extension ID of the extension to validate against.
            href : Optional HREF of the STAC object being validated.
        """
        raise NotImplementedError

    def validate(
        self,
        stac_dict: dict[str, Any],
        stac_object_type: STACObjectType,
        stac_version: str,
        extensions: list[str],
        href: str | None = None,
    ) -> list[Any]:
        """Validate a STAC object JSON.

        Args:
            stac_dict : Dictionary that is the STAC JSON of the object.
            stac_object_type : The stac object type of the object encoded in
                stac_dict. One of :class:`~pystac.STACObjectType`.
            stac_version : The version of STAC to validate the object against.
            extensions : Extension IDs for this stac object.
            href : Optional href of the STAC object being validated.

        Returns:
            List[Any]: List of return values from the validation calls for the
               core object and any extensions. Element type is specific to the
               STACValidator implementation.
        """
        results: list[Any] = []

        # Pass the dict through JSON serialization and parsing, otherwise
        # some valid properties can be marked as invalid (e.g. tuples in
        # coordinate sequences for geometries).
        json_dict = json.loads(json.dumps(stac_dict))
        core_result = self.validate_core(
            json_dict, stac_object_type, stac_version, href
        )
        if core_result is not None:
            results.append(core_result)

        for extension_id in extensions:
            ext_result = self.validate_extension(
                json_dict, stac_object_type, stac_version, extension_id, href
            )
            if ext_result is not None:
                results.append(ext_result)

        return results


class JsonSchemaSTACValidator(STACValidator):
    """Validate STAC based on JSON Schemas.

    This validator uses JSON schemas, read from URIs provided by a
    :class:`~pystac.validation.schema_uri_map.SchemaUriMap`, to validate STAC core
    objects and extensions.

    Args:
        schema_uri_map : The :class:`~pystac.validation.schema_uri_map.SchemaUriMap`
            that defines where
            the validator will retrieve the JSON schemas for validation.
            Defaults to an instance of
            :class:`~pystac.validation.schema_uri_map.DefaultSchemaUriMap`

    Note:
    This class requires the ``jsonschema`` library to be installed.
    """

    schema_uri_map: SchemaUriMap
    schema_cache: dict[str, dict[str, Any]]

    def __init__(self, schema_uri_map: SchemaUriMap | None = None) -> None:
        if not HAS_JSONSCHEMA:
            raise ImportError("Cannot instantiate, requires jsonschema package")

        if schema_uri_map is not None:
            self.schema_uri_map = schema_uri_map
        else:
            self.schema_uri_map = DefaultSchemaUriMap()

        self.schema_cache = get_local_schema_cache()

    def _get_schema(self, schema_uri: str) -> dict[str, Any]:
        if schema_uri not in self.schema_cache:
            try:
                s = json.loads(pystac.StacIO.default().read_text(schema_uri))
            except Exception as error:
                raise GetSchemaError(schema_uri, error) from error
            self.schema_cache[schema_uri] = s
            id_field = "$id" if "$id" in s else "id"
            if not s[id_field].startswith("http"):
                s[id_field] = schema_uri
        return self.schema_cache[schema_uri]

    @property
    def registry(self) -> Any:
        def retrieve(schema_uri: str) -> Resource[dict[str, Any]]:
            return Resource.from_contents(self._get_schema(schema_uri))

        return Registry(retrieve=retrieve).with_resources(  # type: ignore
            [(k, Resource.from_contents(v)) for k, v in self.schema_cache.items()]
        )

    def get_schema_from_uri(self, schema_uri: str) -> tuple[dict[str, Any], Any]:
        """DEPRECATED"""
        warnings.warn(
            "get_schema_from_uri is deprecated and will be removed in v2.",
            DeprecationWarning,
        )
        return self._get_schema(schema_uri), self.registry

    def _validate_from_uri(
        self,
        stac_dict: dict[str, Any],
        stac_object_type: STACObjectType,
        schema_uri: str,
        href: str | None = None,
    ) -> None:
        try:
            schema = self._get_schema(schema_uri)
            # This block is cribbed (w/ change in error handling) from
            # jsonschema.validate
            cls = jsonschema.validators.validator_for(schema)
            cls.check_schema(schema)
            validator = cls(schema, registry=self.registry)
            errors = list(validator.iter_errors(stac_dict))
        except Exception as e:
            logger.error(f"Exception while validating {stac_object_type} href: {href}")
            logger.exception(e)
            raise
        if errors:
            stac_id = stac_dict.get("id", None)
            msg = f"Validation failed for {stac_object_type} "
            if href is not None:
                msg += f"at {href} "
            if stac_id is not None:
                msg += f"with ID {stac_id} "
            msg += f"against schema at {schema_uri}"

            best = jsonschema.exceptions.best_match(errors)
            if best:
                msg += "\n" + str(best)
            raise STACValidationError(msg, source=errors) from best

    def validate_core(
        self,
        stac_dict: dict[str, Any],
        stac_object_type: STACObjectType,
        stac_version: str,
        href: str | None = None,
    ) -> str | None:
        """Validate a core stac object.

        Return value can be None or specific to the implementation.

        Args:
            stac_dict : Dictionary that is the STAC JSON of the object.
            stac_object_type : The stac object type of the object encoded in
                stac_dict. One of :class:`~pystac.STACObjectType`.
            stac_version : The version of STAC to validate the object against.
            href : Optional HREF of the STAC object being validated.

        Returns:
           str: URI for the JSON schema that was validated against, or None if
               no validation occurred.

        Raises:
           STACValidationError:  if stac_dict is not valid. The exception is raised from
               the "best" error, as determined by the jsonschema library. To access all
               jsonschema validation errors, use ``STACValidationError.source``.

        """
        schema_uri = self.schema_uri_map.get_object_schema_uri(
            stac_object_type, stac_version
        )

        if schema_uri is None:
            return None

        self._validate_from_uri(stac_dict, stac_object_type, schema_uri, href=href)

        return schema_uri

    def validate_extension(
        self,
        stac_dict: dict[str, Any],
        stac_object_type: STACObjectType,
        stac_version: str,
        extension_id: str,
        href: str | None = None,
    ) -> str | None:
        """Validate an extension stac object.

        Return value can be None or specific to the implementation.

        Args:
            stac_dict : Dictionary that is the STAC JSON of the object.
            stac_object_type : The stac object type of the object encoded in
                stac_dict. One of :class:`~pystac.STACObjectType`.
            stac_version : The version of STAC to validate the object against.
            extension_id : The extension ID to validate against.
            href : Optional HREF of the STAC object being validated.

        Returns:
           str: URI for the JSON schema that was validated against, or None if
               no validation occurred.

        Raises:
           STACValidationError: if stac_dict is not valid. The exception is raised from
               the "best" error, as determined by the jsonschema library. To access all
               jsonschema validation errors, use ``STACValidationError.source``.

        """
        schema_uri = extension_id

        if schema_uri is None:
            return None

        schema_uri = pystac.utils.make_absolute_href(schema_uri, href)

        self._validate_from_uri(stac_dict, stac_object_type, schema_uri, href)

        return schema_uri



"""Tests for pystac.extensions.grid."""

# This is for the type checking on GridTest.test_clear_code
# mypy: warn_unused_ignores=False

from datetime import datetime
from typing import Any

import pytest

import pystac
from pystac import ExtensionTypeError
from pystac.extensions import grid
from pystac.extensions.grid import GridExtension
from tests.conftest import get_data_file
from tests.utils import TestCases

code = "MGRS-4CFJ"


@pytest.fixture
def sentinel2_example_item() -> pystac.Item:
    return pystac.Item.from_file(
        TestCases.get_path("data-files/grid/example-sentinel2.json")
    )


@pytest.fixture
def item() -> pystac.Item:
    """Create basic test items that are only slightly different."""
    asset_id = "an/asset"
    start = datetime(2018, 1, 2)
    item = pystac.Item(
        id=asset_id, geometry=None, bbox=None, datetime=start, properties={}
    )

    GridExtension.add_to(item)
    return item


def test_stac_extensions(item: pystac.Item) -> None:
    assert GridExtension.has_extension(item)


def test_item_repr(item: pystac.Item) -> None:
    grid_item_ext = GridExtension.ext(item)
    assert f"<ItemGridExtension Item id={item.id}>" == grid_item_ext.__repr__()


@pytest.mark.vcr()
def test_attributes(item: pystac.Item) -> None:
    GridExtension.ext(item).apply(code)
    assert code == GridExtension.ext(item).code
    item.validate()


def test_invalid_code_value(item: pystac.Item) -> None:
    with pytest.raises(ValueError):
        GridExtension.ext(item).apply("not_a_valid_code")


@pytest.mark.vcr()
def test_modify(item: pystac.Item) -> None:
    GridExtension.ext(item).apply(code)
    GridExtension.ext(item).apply(code + "a")
    assert code + "a" == GridExtension.ext(item).code
    item.validate()


def test_from_dict() -> None:
    d: dict[str, Any] = {
        "type": "Feature",
        "stac_version": "1.1.0",
        "id": "an/asset",
        "properties": {
            "grid:code": code,
            "datetime": "2018-01-02T00:00:00Z",
        },
        "geometry": None,
        "links": [],
        "assets": {},
        "stac_extensions": [GridExtension.get_schema_uri()],
    }
    item = pystac.Item.from_dict(d)
    assert code == GridExtension.ext(item).code


def test_to_from_dict(item: pystac.Item) -> None:
    GridExtension.ext(item).apply(code)
    d = item.to_dict()
    assert code == d["properties"][grid.CODE_PROP]

    item = pystac.Item.from_dict(d)
    assert code == GridExtension.ext(item).code


def test_clear_code(item: pystac.Item) -> None:
    GridExtension.ext(item).apply(code)

    with pytest.raises(ValueError):
        # Ignore type errors because this test intentionally checks behavior
        # that does not conform to the type signature.
        # https://github.com/stac-utils/pystac/pull/878#discussion_r957352232
        GridExtension.ext(item).code = None  # type: ignore
    with pytest.raises(ValueError):
        # First segment has to be all caps
        # https://github.com/stac-utils/pystac/pull/878#discussion_r957354927
        GridExtension.ext(item).code = "this-is-not-a-grid-code"
    with pytest.raises(ValueError):
        # Folks might try to put an epsg code in
        # https://github.com/stac-utils/pystac/pull/878#discussion_r957355415
        GridExtension.ext(item).code = "4326"
    with pytest.raises(ValueError):
        # Folks might try to put an epsg code in
        # https://github.com/stac-utils/pystac/pull/878#discussion_r957355415
        GridExtension.ext(item).code = "EPSG:4326"


def test_extension_not_implemented(sentinel2_example_item: pystac.Item) -> None:
    # Should raise exception if Item does not include extension URI
    item = sentinel2_example_item
    item.stac_extensions.remove(GridExtension.get_schema_uri())

    with pytest.raises(pystac.ExtensionNotImplemented):
        _ = GridExtension.ext(item)

    # Should raise exception if owning Item does not include extension URI
    item.properties["grid:code"] = None

    with pytest.raises(pystac.ExtensionNotImplemented):
        _ = GridExtension.ext(item)


def test_item_ext_add_to(sentinel2_example_item: pystac.Item) -> None:
    item = sentinel2_example_item
    item.stac_extensions.remove(GridExtension.get_schema_uri())
    assert GridExtension.get_schema_uri() not in item.stac_extensions

    _ = GridExtension.ext(item, add_if_missing=True)

    assert GridExtension.get_schema_uri() in item.stac_extensions


def test_should_raise_when_passing_invalid_extension_object() -> None:
    with pytest.raises(
        ExtensionTypeError, match=r"^GridExtension does not apply to type 'object'$"
    ):
        # intentionally calling it wrong so tell mypy to ignore this line:
        GridExtension.ext(object())  # type: ignore


@pytest.fixture
def ext_item() -> pystac.Item:
    ext_item_uri = get_data_file("grid/example-sentinel2.json")
    return pystac.Item.from_file(ext_item_uri)


def test_older_extension_version(ext_item: pystac.Item) -> None:
    old = "https://stac-extensions.github.io/grid/v1.0.0/schema.json"
    new = "https://stac-extensions.github.io/grid/v1.1.0/schema.json"

    stac_extensions = set(ext_item.stac_extensions)
    stac_extensions.remove(new)
    stac_extensions.add(old)
    item_as_dict = ext_item.to_dict(include_self_link=False, transform_hrefs=False)
    item_as_dict["stac_extensions"] = list(stac_extensions)
    item = pystac.Item.from_dict(item_as_dict, migrate=False)
    assert GridExtension.has_extension(item)
    assert old in item.stac_extensions

    migrated_item = pystac.Item.from_dict(item_as_dict, migrate=True)
    assert GridExtension.has_extension(migrated_item)
    assert new in migrated_item.stac_extensions



"""Implements the :stac-ext:`Synthetic-Aperture Radar (SAR) Extension <sar>`."""

from __future__ import annotations

from collections.abc import Iterable
from typing import (
    Any,
    Generic,
    Literal,
    TypeVar,
    cast,
)

import pystac
from pystac.extensions.base import (
    ExtensionManagementMixin,
    PropertiesExtension,
    SummariesExtension,
)
from pystac.extensions.hooks import ExtensionHooks
from pystac.serialization.identify import STACJSONDescription, STACVersionID
from pystac.summaries import RangeSummary
from pystac.utils import StringEnum, get_required, map_opt

#: Generalized version of :class:`~pystac.Item`, :class:`~pystac.Asset` or
#: :class:`~pystac.ItemAssetDefinition`
T = TypeVar("T", pystac.Item, pystac.Asset, pystac.ItemAssetDefinition)

SCHEMA_URI: str = "https://stac-extensions.github.io/sar/v1.0.0/schema.json"
PREFIX: str = "sar:"

# Required
INSTRUMENT_MODE_PROP: str = PREFIX + "instrument_mode"
FREQUENCY_BAND_PROP: str = PREFIX + "frequency_band"
POLARIZATIONS_PROP: str = PREFIX + "polarizations"
PRODUCT_TYPE_PROP: str = PREFIX + "product_type"

# Not required
CENTER_FREQUENCY_PROP: str = PREFIX + "center_frequency"
RESOLUTION_RANGE_PROP: str = PREFIX + "resolution_range"
RESOLUTION_AZIMUTH_PROP: str = PREFIX + "resolution_azimuth"
PIXEL_SPACING_RANGE_PROP: str = PREFIX + "pixel_spacing_range"
PIXEL_SPACING_AZIMUTH_PROP: str = PREFIX + "pixel_spacing_azimuth"
LOOKS_RANGE_PROP: str = PREFIX + "looks_range"
LOOKS_AZIMUTH_PROP: str = PREFIX + "looks_azimuth"
LOOKS_EQUIVALENT_NUMBER_PROP: str = PREFIX + "looks_equivalent_number"
OBSERVATION_DIRECTION_PROP: str = PREFIX + "observation_direction"


class FrequencyBand(StringEnum):
    P = "P"
    L = "L"
    S = "S"
    C = "C"
    X = "X"
    KU = "Ku"
    K = "K"
    KA = "Ka"


class Polarization(StringEnum):
    HH = "HH"
    VV = "VV"
    HV = "HV"
    VH = "VH"


class ObservationDirection(StringEnum):
    LEFT = "left"
    RIGHT = "right"


class SarExtension(
    Generic[T],
    PropertiesExtension,
    ExtensionManagementMixin[pystac.Item | pystac.Collection],
):
    """An abstract class that can be used to extend the properties of an
    :class:`~pystac.Item` or :class:`~pystac.Asset` with properties from the
    :stac-ext:`SAR Extension <sar>`. This class is generic over the type of
    STAC Object to be extended (e.g. :class:`~pystac.Item`,
    :class:`~pystac.Asset`).

    To create a concrete instance of :class:`SarExtension`, use the
    :meth:`SarExtension.ext` method. For example:

    .. code-block:: python

       >>> item: pystac.Item = ...
       >>> sar_ext = SarExtension.ext(item)
    """

    name: Literal["sar"] = "sar"

    def apply(
        self,
        instrument_mode: str,
        frequency_band: FrequencyBand,
        polarizations: list[Polarization],
        product_type: str,
        center_frequency: float | None = None,
        resolution_range: float | None = None,
        resolution_azimuth: float | None = None,
        pixel_spacing_range: float | None = None,
        pixel_spacing_azimuth: float | None = None,
        looks_range: int | None = None,
        looks_azimuth: int | None = None,
        looks_equivalent_number: float | None = None,
        observation_direction: ObservationDirection | None = None,
    ) -> None:
        """Applies sar extension properties to the extended Item.

        Args:
            instrument_mode : The name of the sensor acquisition mode that is
                commonly used. This should be the short name, if available. For example,
                WV for "Wave mode."
            frequency_band : The common name for the frequency band to
                make it easier to search for bands across instruments. See section
                "Common Frequency Band Names" for a list of accepted names.
            polarizations : Any combination of polarizations.
            product_type : The product type, for example SSC, MGD, or SGC.
            center_frequency : Optional center frequency of the instrument in
                gigahertz (GHz).
            resolution_range : Optional range resolution, which is the maximum
                ability to distinguish two adjacent targets perpendicular to the flight
                path, in meters (m).
            resolution_azimuth : Optional azimuth resolution, which is the
                maximum ability to distinguish two adjacent targets parallel to the
                flight path, in meters (m).
            pixel_spacing_range : Optional range pixel spacing, which is the
                distance between adjacent pixels perpendicular to the flight path,
                in meters (m). Strongly RECOMMENDED to be specified for
                products of type GRD.
            pixel_spacing_azimuth : Optional azimuth pixel spacing, which is the
                distance between adjacent pixels parallel to the flight path, in
                meters (m). Strongly RECOMMENDED to be specified for products of
                type GRD.
            looks_range : Optional number of groups of signal samples (looks)
                perpendicular to the flight path.
            looks_azimuth : Optional number of groups of signal samples (looks)
                parallel to the flight path.
            looks_equivalent_number : Optional equivalent number of looks (ENL).
            observation_direction : Optional Antenna pointing
                direction relative to the flight trajectory of the satellite.
        """
        self.instrument_mode = instrument_mode
        self.frequency_band = frequency_band
        self.polarizations = polarizations
        self.product_type = product_type
        if center_frequency:
            self.center_frequency = center_frequency
        if resolution_range:
            self.resolution_range = resolution_range
        if resolution_azimuth:
            self.resolution_azimuth = resolution_azimuth
        if pixel_spacing_range:
            self.pixel_spacing_range = pixel_spacing_range
        if pixel_spacing_azimuth:
            self.pixel_spacing_azimuth = pixel_spacing_azimuth
        if looks_range:
            self.looks_range = looks_range
        if looks_azimuth:
            self.looks_azimuth = looks_azimuth
        if looks_equivalent_number:
            self.looks_equivalent_number = looks_equivalent_number
        if observation_direction:
            self.observation_direction = observation_direction

    @property
    def instrument_mode(self) -> str:
        """Gets or sets an instrument mode string for the item."""
        return get_required(
            self._get_property(INSTRUMENT_MODE_PROP, str), self, INSTRUMENT_MODE_PROP
        )

    @instrument_mode.setter
    def instrument_mode(self, v: str) -> None:
        self._set_property(INSTRUMENT_MODE_PROP, v, pop_if_none=False)

    @property
    def frequency_band(self) -> FrequencyBand:
        """Gets or sets a FrequencyBand for the item."""
        return get_required(
            map_opt(
                lambda x: FrequencyBand(x), self._get_property(FREQUENCY_BAND_PROP, str)
            ),
            self,
            FREQUENCY_BAND_PROP,
        )

    @frequency_band.setter
    def frequency_band(self, v: FrequencyBand) -> None:
        self._set_property(FREQUENCY_BAND_PROP, v.value, pop_if_none=False)

    @property
    def polarizations(self) -> list[Polarization]:
        """Gets or sets a list of polarizations for the item."""
        return get_required(
            map_opt(
                lambda values: [Polarization(v) for v in values],
                self._get_property(POLARIZATIONS_PROP, list[str]),
            ),
            self,
            POLARIZATIONS_PROP,
        )

    @polarizations.setter
    def polarizations(self, values: list[Polarization]) -> None:
        if not isinstance(values, list):
            raise pystac.STACError(f'polarizations must be a list. Invalid "{values}"')
        self._set_property(
            POLARIZATIONS_PROP, [v.value for v in values], pop_if_none=False
        )

    @property
    def product_type(self) -> str:
        """Gets or sets a product type string for the item."""
        return get_required(
            self._get_property(PRODUCT_TYPE_PROP, str), self, PRODUCT_TYPE_PROP
        )

    @product_type.setter
    def product_type(self, v: str) -> None:
        self._set_property(PRODUCT_TYPE_PROP, v, pop_if_none=False)

    @property
    def center_frequency(self) -> float | None:
        """Gets or sets a center frequency for the item."""
        return self._get_property(CENTER_FREQUENCY_PROP, float)

    @center_frequency.setter
    def center_frequency(self, v: float | None) -> None:
        self._set_property(CENTER_FREQUENCY_PROP, v)

    @property
    def resolution_range(self) -> float | None:
        """Gets or sets a resolution range for the item."""
        return self._get_property(RESOLUTION_RANGE_PROP, float)

    @resolution_range.setter
    def resolution_range(self, v: float | None) -> None:
        self._set_property(RESOLUTION_RANGE_PROP, v)

    @property
    def resolution_azimuth(self) -> float | None:
        """Gets or sets a resolution azimuth for the item."""
        return self._get_property(RESOLUTION_AZIMUTH_PROP, float)

    @resolution_azimuth.setter
    def resolution_azimuth(self, v: float | None) -> None:
        self._set_property(RESOLUTION_AZIMUTH_PROP, v)

    @property
    def pixel_spacing_range(self) -> float | None:
        """Gets or sets a pixel spacing range for the item."""
        return self._get_property(PIXEL_SPACING_RANGE_PROP, float)

    @pixel_spacing_range.setter
    def pixel_spacing_range(self, v: float | None) -> None:
        self._set_property(PIXEL_SPACING_RANGE_PROP, v)

    @property
    def pixel_spacing_azimuth(self) -> float | None:
        """Gets or sets a pixel spacing azimuth for the item."""
        return self._get_property(PIXEL_SPACING_AZIMUTH_PROP, float)

    @pixel_spacing_azimuth.setter
    def pixel_spacing_azimuth(self, v: float | None) -> None:
        self._set_property(PIXEL_SPACING_AZIMUTH_PROP, v)

    @property
    def looks_range(self) -> int | None:
        """Gets or sets a looks range for the item."""
        return self._get_property(LOOKS_RANGE_PROP, int)

    @looks_range.setter
    def looks_range(self, v: int | None) -> None:
        self._set_property(LOOKS_RANGE_PROP, v)

    @property
    def looks_azimuth(self) -> int | None:
        """Gets or sets a looks azimuth for the item."""
        return self._get_property(LOOKS_AZIMUTH_PROP, int)

    @looks_azimuth.setter
    def looks_azimuth(self, v: int | None) -> None:
        self._set_property(LOOKS_AZIMUTH_PROP, v)

    @property
    def looks_equivalent_number(self) -> float | None:
        """Gets or sets a looks equivalent number for the item."""
        return self._get_property(LOOKS_EQUIVALENT_NUMBER_PROP, float)

    @looks_equivalent_number.setter
    def looks_equivalent_number(self, v: float | None) -> None:
        self._set_property(LOOKS_EQUIVALENT_NUMBER_PROP, v)

    @property
    def observation_direction(self) -> ObservationDirection | None:
        """Gets or sets an observation direction for the item."""
        return map_opt(
            ObservationDirection, self._get_property(OBSERVATION_DIRECTION_PROP, str)
        )

    @observation_direction.setter
    def observation_direction(self, v: ObservationDirection | None) -> None:
        self._set_property(OBSERVATION_DIRECTION_PROP, map_opt(lambda x: x.value, v))

    @classmethod
    def get_schema_uri(cls) -> str:
        return SCHEMA_URI

    @classmethod
    def ext(cls, obj: T, add_if_missing: bool = False) -> SarExtension[T]:
        """Extends the given STAC Object with properties from the :stac-ext:`SAR
        Extension <sar>`.

        This extension can be applied to instances of :class:`~pystac.Item` or
        :class:`~pystac.Asset`.

        Raises:

            pystac.ExtensionTypeError : If an invalid object type is passed.
        """
        if isinstance(obj, pystac.Item):
            cls.ensure_has_extension(obj, add_if_missing)
            return cast(SarExtension[T], ItemSarExtension(obj))
        elif isinstance(obj, pystac.Asset):
            if obj.owner is not None and not isinstance(obj.owner, pystac.Item):
                raise pystac.ExtensionTypeError(
                    "SAR extension does not apply to Collection Assets."
                )
            cls.ensure_owner_has_extension(obj, add_if_missing)
            return cast(SarExtension[T], AssetSarExtension(obj))
        elif isinstance(obj, pystac.ItemAssetDefinition):
            cls.ensure_owner_has_extension(obj, add_if_missing)
            return cast(SarExtension[T], ItemAssetsSarExtension(obj))
        else:
            raise pystac.ExtensionTypeError(cls._ext_error_message(obj))

    @classmethod
    def summaries(
        cls, obj: pystac.Collection, add_if_missing: bool = False
    ) -> SummariesSarExtension:
        """Returns the extended summaries object for the given collection."""
        cls.ensure_has_extension(obj, add_if_missing)
        return SummariesSarExtension(obj)


class ItemSarExtension(SarExtension[pystac.Item]):
    """A concrete implementation of :class:`SarExtension` on an :class:`~pystac.Item`
    that extends the properties of the Item to include properties defined in the
    :stac-ext:`SAR Extension <sar>`.

    This class should generally not be instantiated directly. Instead, call
    :meth:`SarExtension.ext` on an :class:`~pystac.Item` to extend it.
    """

    item: pystac.Item
    """The :class:`~pystac.Item` being extended."""

    properties: dict[str, Any]
    """The :class:`~pystac.Item` properties, including extension properties."""

    def __init__(self, item: pystac.Item):
        self.item = item
        self.properties = item.properties

    def __repr__(self) -> str:
        return f"<ItemSarExtension Item id={self.item.id}>"


class AssetSarExtension(SarExtension[pystac.Asset]):
    """A concrete implementation of :class:`SarExtension` on an :class:`~pystac.Asset`
    that extends the Asset fields to include properties defined in the
    :stac-ext:`SAR Extension <sar>`.

    This class should generally not be instantiated directly. Instead, call
    :meth:`SarExtension.ext` on an :class:`~pystac.Asset` to extend it.
    """

    asset_href: str
    """The ``href`` value of the :class:`~pystac.Asset` being extended."""

    properties: dict[str, Any]
    """The :class:`~pystac.Asset` fields, including extension properties."""

    additional_read_properties: Iterable[dict[str, Any]] | None = None
    """If present, this will be a list containing 1 dictionary representing the
    properties of the owning :class:`~pystac.Item`."""

    def __init__(self, asset: pystac.Asset):
        self.asset_href = asset.href
        self.properties = asset.extra_fields
        if asset.owner and isinstance(asset.owner, pystac.Item):
            self.additional_read_properties = [asset.owner.properties]

    def __repr__(self) -> str:
        return f"<AssetSarExtension Asset href={self.asset_href}>"


class ItemAssetsSarExtension(SarExtension[pystac.ItemAssetDefinition]):
    properties: dict[str, Any]
    asset_defn: pystac.ItemAssetDefinition

    def __init__(self, item_asset: pystac.ItemAssetDefinition):
        self.asset_defn = item_asset
        self.properties = item_asset.properties


class SummariesSarExtension(SummariesExtension):
    """A concrete implementation of :class:`~pystac.extensions.base.SummariesExtension`
    that extends the ``summaries`` field of a :class:`~pystac.Collection` to include
    properties defined in the :stac-ext:`SAR Extension <sar>`.
    """

    @property
    def instrument_mode(self) -> list[str] | None:
        """Get or sets the summary of :attr:`SarExtension.instrument_mode` values
        for this Collection.
        """

        return self.summaries.get_list(INSTRUMENT_MODE_PROP)

    @instrument_mode.setter
    def instrument_mode(self, v: list[str] | None) -> None:
        self._set_summary(INSTRUMENT_MODE_PROP, v)

    @property
    def frequency_band(self) -> list[FrequencyBand] | None:
        """Get or sets the summary of :attr:`SarExtension.frequency_band` values
        for this Collection.
        """

        return self.summaries.get_list(FREQUENCY_BAND_PROP)

    @frequency_band.setter
    def frequency_band(self, v: list[FrequencyBand] | None) -> None:
        self._set_summary(FREQUENCY_BAND_PROP, v)

    @property
    def center_frequency(self) -> RangeSummary[float] | None:
        """Get or sets the summary of :attr:`SarExtension.center_frequency` values
        for this Collection.
        """

        return self.summaries.get_range(CENTER_FREQUENCY_PROP)

    @center_frequency.setter
    def center_frequency(self, v: RangeSummary[float] | None) -> None:
        self._set_summary(CENTER_FREQUENCY_PROP, v)

    @property
    def polarizations(self) -> list[Polarization] | None:
        """Get or sets the summary of :attr:`SarExtension.polarizations` values
        for this Collection.
        """

        return self.summaries.get_list(POLARIZATIONS_PROP)

    @polarizations.setter
    def polarizations(self, v: list[Polarization] | None) -> None:
        self._set_summary(POLARIZATIONS_PROP, v)

    @property
    def product_type(self) -> list[str] | None:
        """Get or sets the summary of :attr:`SarExtension.product_type` values
        for this Collection.
        """

        return self.summaries.get_list(PRODUCT_TYPE_PROP)

    @product_type.setter
    def product_type(self, v: list[str] | None) -> None:
        self._set_summary(PRODUCT_TYPE_PROP, v)

    @property
    def resolution_range(self) -> RangeSummary[float] | None:
        """Get or sets the summary of :attr:`SarExtension.resolution_range` values
        for this Collection.
        """

        return self.summaries.get_range(RESOLUTION_RANGE_PROP)

    @resolution_range.setter
    def resolution_range(self, v: RangeSummary[float] | None) -> None:
        self._set_summary(RESOLUTION_RANGE_PROP, v)

    @property
    def resolution_azimuth(self) -> RangeSummary[float] | None:
        """Get or sets the summary of :attr:`SarExtension.resolution_azimuth` values
        for this Collection.
        """

        return self.summaries.get_range(RESOLUTION_AZIMUTH_PROP)

    @resolution_azimuth.setter
    def resolution_azimuth(self, v: RangeSummary[float] | None) -> None:
        self._set_summary(RESOLUTION_AZIMUTH_PROP, v)

    @property
    def pixel_spacing_range(self) -> RangeSummary[float] | None:
        """Get or sets the summary of :attr:`SarExtension.pixel_spacing_range` values
        for this Collection.
        """

        return self.summaries.get_range(PIXEL_SPACING_RANGE_PROP)

    @pixel_spacing_range.setter
    def pixel_spacing_range(self, v: RangeSummary[float] | None) -> None:
        self._set_summary(PIXEL_SPACING_RANGE_PROP, v)

    @property
    def pixel_spacing_azimuth(self) -> RangeSummary[float] | None:
        """Get or sets the summary of :attr:`SarExtension.pixel_spacing_azimuth` values
        for this Collection.
        """

        return self.summaries.get_range(PIXEL_SPACING_AZIMUTH_PROP)

    @pixel_spacing_azimuth.setter
    def pixel_spacing_azimuth(self, v: RangeSummary[float] | None) -> None:
        self._set_summary(PIXEL_SPACING_AZIMUTH_PROP, v)

    @property
    def looks_range(self) -> RangeSummary[int] | None:
        """Get or sets the summary of :attr:`SarExtension.looks_range` values
        for this Collection.
        """

        return self.summaries.get_range(LOOKS_RANGE_PROP)

    @looks_range.setter
    def looks_range(self, v: RangeSummary[int] | None) -> None:
        self._set_summary(LOOKS_RANGE_PROP, v)

    @property
    def looks_azimuth(self) -> RangeSummary[int] | None:
        """Get or sets the summary of :attr:`SarExtension.looks_azimuth` values
        for this Collection.
        """

        return self.summaries.get_range(LOOKS_AZIMUTH_PROP)

    @looks_azimuth.setter
    def looks_azimuth(self, v: RangeSummary[int] | None) -> None:
        self._set_summary(LOOKS_AZIMUTH_PROP, v)

    @property
    def looks_equivalent_number(self) -> RangeSummary[float] | None:
        """Get or sets the summary of :attr:`SarExtension.looks_equivalent_number`
        values for this Collection.
        """

        return self.summaries.get_range(LOOKS_EQUIVALENT_NUMBER_PROP)

    @looks_equivalent_number.setter
    def looks_equivalent_number(self, v: RangeSummary[float] | None) -> None:
        self._set_summary(LOOKS_EQUIVALENT_NUMBER_PROP, v)

    @property
    def observation_direction(self) -> list[ObservationDirection] | None:
        """Get or sets the summary of :attr:`SarExtension.observation_direction` values
        for this Collection.
        """

        return self.summaries.get_list(OBSERVATION_DIRECTION_PROP)

    @observation_direction.setter
    def observation_direction(self, v: list[ObservationDirection] | None) -> None:
        self._set_summary(OBSERVATION_DIRECTION_PROP, v)


class SarExtensionHooks(ExtensionHooks):
    schema_uri = SCHEMA_URI
    prev_extension_ids = {"sar"}
    stac_object_types = {pystac.STACObjectType.ITEM}

    def migrate(
        self, obj: dict[str, Any], version: STACVersionID, info: STACJSONDescription
    ) -> None:
        if version < "0.9":
            # Some sar fields became common_metadata
            if (
                PREFIX + "platform" in obj["properties"]
                and "platform" not in obj["properties"]
            ):
                obj["properties"]["platform"] = obj["properties"].pop(
                    PREFIX + "platform"
                )

            if (
                PREFIX + "instrument" in obj["properties"]
                and "instruments" not in obj["properties"]
            ):
                obj["properties"]["instruments"] = [
                    obj["properties"].pop(PREFIX + "instrument")
                ]

            if (
                PREFIX + "constellation" in obj["properties"]
                and "constellation" not in obj["properties"]
            ):
                obj["properties"]["constellation"] = obj["properties"].pop(
                    PREFIX + "constellation"
                )

            # Some SAR fields changed property names
            if (
                PREFIX + "type" in obj["properties"]
                and PRODUCT_TYPE_PROP not in obj["properties"]
            ):
                obj["properties"][PRODUCT_TYPE_PROP] = obj["properties"].pop(
                    PREFIX + "type"
                )

            if (
                PREFIX + "polarization" in obj["properties"]
                and POLARIZATIONS_PROP not in obj["properties"]
            ):
                obj["properties"][POLARIZATIONS_PROP] = [
                    obj["properties"].pop(PREFIX + "polarization")
                ]

        super().migrate(obj, version, info)


SAR_EXTENSION_HOOKS: ExtensionHooks = SarExtensionHooks()



from __future__ import annotations

import posixpath
from abc import ABC, abstractmethod
from collections import OrderedDict
from collections.abc import Callable
from typing import TYPE_CHECKING, Any

import pystac
from pystac.utils import is_file_path

if TYPE_CHECKING:
    from pystac.catalog import Catalog
    from pystac.collection import Collection
    from pystac.item import Item
    from pystac.stac_object import STACObject


class TemplateError(Exception):
    """DEPRECATED.

    .. deprecated:: 1.7.0
        Use :class:`pystac.errors.TemplateError` instead.

    Exception thrown when an error occurs during converting a template
    string into data for :class:`~pystac.layout.LayoutTemplate`
    """

    def __init__(self, *args, **kwargs) -> None:  # type: ignore
        import warnings

        warnings.warn(
            message=(
                "TemplateError in pystac.layout is deprecated and will be "
                "removed in pystac version 2.0.0. Use TemplateError in "
                "pystac.errors instead."
            ),
            category=DeprecationWarning,
        )
        super().__init__(*args, **kwargs)


class LayoutTemplate:
    """Represents a template that can be used for deriving paths or other information
    based on properties of STAC objects supplied as a template string.

    Template strings use variables that can be pulled out of Catalogs, Collections
    and Items. The variable names can represent properties on the object itself,
    or keys in the dictionaries of the properties or extra_fields on the
    object (if they exist). The search order is as follows:

    - The object's attributes
    - Keys in the ``properties`` attribute, if it exists.
    - Keys in the ``extra_fields`` attribute, if it exists.

    Some special keys can be used in template variables:

    +--------------------+--------------------------------------------------------+
    | Template variable  | Meaning                                                |
    +====================+========================================================+
    | ``year``           | The year of an Item's datetime, or                     |
    |                    | start_datetime if datetime is null                     |
    +--------------------+--------------------------------------------------------+
    | ``month``          | The month of an Item's datetime, or                    |
    |                    | start_datetime if datetime is null                     |
    +--------------------+--------------------------------------------------------+
    | ``day``            | The day of an Item's datetime, or                      |
    |                    | start_datetime if datetime is null                     |
    +--------------------+--------------------------------------------------------+
    | ``date``           | The date (iso format) of an Item's                     |
    |                    | datetime, or start_datetime if datetime is null        |
    +--------------------+--------------------------------------------------------+
    | ``collection``     | The collection ID of an Item's collection.             |
    +--------------------+--------------------------------------------------------+

    The forward slash (``/``) should be used as path separator in the template
    string regardless of the system path separator (thus both in POSIX-compliant
    and Windows environments).

    Examples::

        # Uses the year, month and day of the item
        template = LayoutTemplate("${year}/${month}/${day}")

        # Uses a custom extension properties found on in the item properties.
        template = LayoutTemplate("${landsat:path}/${landsat:row}")

        # Uses the collection ID and a common metadata property for an item.
        template = LayoutTemplate("${collection}/${common_metadata.license}")

    Args:
        template : The template string to use.
        defaults : A dictionary of template vars to values. These values
            will be used in case a value cannot be derived from a stac object.
    """

    template: str
    """The template string to use."""

    defaults: dict[str, str]
    """A dictionary of template vars to values. These values will be used in case a
    value cannot be derived from a stac object."""

    template_vars: list[str]
    """List of template vars to use when templating."""

    # Special template vars specific to Items
    ITEM_TEMPLATE_VARS = ["date", "year", "month", "day", "collection"]

    def __init__(self, template: str, defaults: dict[str, str] | None = None) -> None:
        from string import Formatter

        self.template = template
        self.defaults = defaults or {}

        # Generate list of template vars
        template_vars = []
        for formatter_parse_result in Formatter().parse(template):
            v = formatter_parse_result[1]
            if v is not None:
                if formatter_parse_result[2] != "":
                    v = f"{v}:{formatter_parse_result[2]}"
                template_vars.append(v)
        self.template_vars = template_vars

    def _get_template_value(self, stac_object: STACObject, template_var: str) -> Any:
        if template_var in self.ITEM_TEMPLATE_VARS:
            if isinstance(stac_object, pystac.Item):
                # Datetime
                dt = stac_object.datetime
                if dt is None:
                    dt = stac_object.common_metadata.start_datetime
                if dt is None:
                    raise pystac.TemplateError(
                        "Item {} does not have a datetime or "
                        "datetime range set; cannot template {} in {}".format(
                            stac_object, template_var, self.template
                        )
                    )

                if template_var == "year":
                    return dt.year
                if template_var == "month":
                    return dt.month
                if template_var == "day":
                    return dt.day
                if template_var == "date":
                    return dt.date().isoformat()

                # Collection
                if template_var == "collection":
                    if stac_object.collection_id is not None:
                        return stac_object.collection_id
                    raise pystac.TemplateError(
                        f"Item {stac_object} does not have a collection ID set; "
                        f"cannot template {template_var} in {self.template}"
                    )
            else:
                raise pystac.TemplateError(
                    '"{}" cannot be used to template non-Item {} in {}'.format(
                        template_var, stac_object, self.template
                    )
                )

        # Allow dot-notation properties for arbitrary object values.
        props = template_var.split(".")
        prop_source: pystac.STACObject | dict[str, Any] | None = None
        error = pystac.TemplateError(
            "Cannot find property {} on {} for template {}".format(
                template_var, stac_object, self.template
            )
        )

        try:
            if hasattr(stac_object, props[0]):
                prop_source = stac_object

            if prop_source is None and hasattr(stac_object, "properties"):
                obj_props: dict[str, Any] | None = stac_object.properties
                if obj_props is not None and props[0] in obj_props:
                    prop_source = obj_props

            if prop_source is None and hasattr(stac_object, "extra_fields"):
                extra_fields: dict[str, Any] | None = stac_object.extra_fields
                if extra_fields is not None and props[0] in extra_fields:
                    prop_source = extra_fields

            if prop_source is None:
                raise error

            v: Any = prop_source
            for prop in template_var.split("."):
                if isinstance(v, dict):
                    if prop not in v:
                        raise error
                    v = v[prop]
                else:
                    if not hasattr(v, prop):
                        raise error
                    v = getattr(v, prop)
        except pystac.TemplateError as e:
            if template_var in self.defaults:
                return self.defaults[template_var]
            raise e

        return v

    def get_template_values(self, stac_object: STACObject) -> dict[str, Any]:
        """Gets a dictionary of template variables to values derived from
        the given stac_object. If the template vars cannot be found in the
        stac object, and defaults was supplied to this template, a default
        value is used; if there is no default then this will raise an error.

        Args:
            stac_object : The STACObject to derive template
                variable values from.

        Returns:
            [dict]: A dictionary with keys being the template variables
            and values being the respective values based on the given
            stac object.

        Raises:
            pystac.TemplateError: If a value for a template variable cannot be
                derived from the stac object and there is no default,
                this error will be raised.
        """
        return OrderedDict(
            [(k, self._get_template_value(stac_object, k)) for k in self.template_vars]
        )

    def substitute(self, stac_object: STACObject) -> str:
        """Substitutes the values derived from
        :meth:`~pystac.layout.LayoutTemplate.get_template_values` into
        the template string for this template.

        Args:
            stac_object : The STACObject to derive template
                variable values from.

        Returns:
            str: The original template supplied to this LayoutTemplate
            with template variables replaced by the values derived
            from this stac object.

        Raises:
            pystac.TemplateError: If a value for a template variable cannot be
                derived from the stac object and there is no default,
                this error will be raised.
        """
        parts = self.get_template_values(stac_object)

        s = self.template
        for key, value in parts.items():
            s = s.replace("${" + f"{key}" + "}", f"{value}")
        return s


class HrefLayoutStrategy(ABC):
    """Base class for HREF Layout strategies."""

    def get_href(
        self, stac_object: STACObject, parent_dir: str, is_root: bool = False
    ) -> str:
        import os

        if is_file_path(parent_dir):
            parent_dir = os.path.dirname(parent_dir)

        if isinstance(stac_object, pystac.Item):
            return self.get_item_href(stac_object, parent_dir)
        elif isinstance(stac_object, pystac.Collection):
            return self.get_collection_href(stac_object, parent_dir, is_root)
        elif isinstance(stac_object, pystac.Catalog):
            return self.get_catalog_href(stac_object, parent_dir, is_root)
        else:
            raise pystac.STACError(f"Unknown STAC object type {stac_object}")

    @abstractmethod
    def get_catalog_href(self, cat: Catalog, parent_dir: str, is_root: bool) -> str:
        raise NotImplementedError

    @abstractmethod
    def get_collection_href(
        self, col: Collection, parent_dir: str, is_root: bool
    ) -> str:
        raise NotImplementedError

    @abstractmethod
    def get_item_href(self, item: Item, parent_dir: str) -> str:
        raise NotImplementedError


class CustomLayoutStrategy(HrefLayoutStrategy):
    """Layout strategy that allows users to supply functions to dictate
    stac object paths.

    Args:
        catalog_func : A function that takes
            a catalog, a parent directory, and a flag specifying whether
            or not this catalog is the root. If it is the root, it is usually
            best to not create a subdirectory and put the Catalog file directly
            in the parent directory. Must return the string path.
        collection_func : A function that
            is used for collections in the same manner at ``catalog_func``. This
            takes the same parameters.
        item_func  : A function that takes
            an item and a parent directory and returns the path to be used
            for the item.
        fallback_strategy : The fallback strategy to
            use if a function is not provided for a stac object type. Defaults to
            :class:`~pystac.layout.BestPracticesLayoutStrategy`
    """

    catalog_func: Callable[[Catalog, str, bool], str] | None
    """A function that takes a :class:`~pystac.Catalog`, a parent directory, and a
    boolean specifying whether or not this Catalog is the root. If it is the root, it
    is usually best to not create a subdirectory and put the Catalog file directly
    in the parent directory. Must return the string path."""

    collection_func: Callable[[Collection, str, bool], str] | None
    """A function that is used for collections in the same manner as
    :attr:`~catalog_func`. This takes the same parameters."""

    fallback_strategy: HrefLayoutStrategy
    """The fallback strategy to use if a function is not provided for a stac object
    type. Defaults to :class:`~pystac.layout.BestPracticesLayoutStrategy`."""

    item_func: Callable[[Item, str], str] | None
    """An optional function that takes an :class:`~pystac.Item` and a parent directory
    and returns the path to be used for the Item."""

    def __init__(
        self,
        catalog_func: Callable[[Catalog, str, bool], str] | None = None,
        collection_func: Callable[[Collection, str, bool], str] | None = None,
        item_func: Callable[[Item, str], str] | None = None,
        fallback_strategy: HrefLayoutStrategy | None = None,
    ):
        self.item_func = item_func
        self.collection_func = collection_func
        self.catalog_func = catalog_func
        if fallback_strategy is None:
            fallback_strategy = BestPracticesLayoutStrategy()
        self.fallback_strategy = fallback_strategy

    def get_catalog_href(self, cat: Catalog, parent_dir: str, is_root: bool) -> str:
        if self.catalog_func is not None:
            result = self.catalog_func(cat, parent_dir, is_root)
            if result is not None:
                return result
        return self.fallback_strategy.get_catalog_href(cat, parent_dir, is_root)

    def get_collection_href(
        self, col: Collection, parent_dir: str, is_root: bool
    ) -> str:
        if self.collection_func is not None:
            result = self.collection_func(col, parent_dir, is_root)
            if result is not None:
                return result
        return self.fallback_strategy.get_collection_href(col, parent_dir, is_root)

    def get_item_href(self, item: Item, parent_dir: str) -> str:
        if self.item_func is not None:
            result = self.item_func(item, parent_dir)
            if result is not None:
                return result
        return self.fallback_strategy.get_item_href(item, parent_dir)


class TemplateLayoutStrategy(HrefLayoutStrategy):
    """Layout strategy that can take strings to be supplied to
    :class:`~pystac.layout.LayoutTemplate` s to derive paths. Template strings
    can be supplied for catalogs, collections and items separately. If no
    template is supplied, a fallback layout strategy is supplied, which defaults
    to the :class:`~pystac.layout.BestPracticesLayoutStrategy`.

    All templated paths will be joined with the parent directory of the stac
    object.

    Args:
        catalog_template : The template string to use for catalog paths.
            Must be a valid template string that can be used by
            :class:`~pystac.layout.LayoutTemplate`
        collection_template : The template string to use for collection paths.
            Must be a valid template string that can be used by
            :class:`~pystac.layout.LayoutTemplate`
        item_template : The template string to use for item paths.
            Must be a valid template string that can be used by
            :class:`~pystac.layout.LayoutTemplate`
        fallback_strategy : The fallback strategy to
            use if a template is not provided. Defaults to
            :class:`~pystac.layout.BestPracticesLayoutStrategy`
    """

    catalog_template: LayoutTemplate | None
    """The template string to use for catalog paths. Must be a valid template string
    that can be used by :class:`~pystac.layout.LayoutTemplate`."""

    collection_template: LayoutTemplate | None
    """The template string to use for collection paths. Must be a valid template string
    that can be used by :class:`~pystac.layout.LayoutTemplate`."""

    fallback_strategy: HrefLayoutStrategy
    """The fallback strategy to use if a template is not provided. Defaults to
    :class:`~pystac.layout.BestPracticesLayoutStrategy`."""

    item_template: LayoutTemplate | None
    """The template string to use for item paths. Must be a valid template string that
    can be used by :class:`~pystac.layout.LayoutTemplate`."""

    def __init__(
        self,
        catalog_template: str | None = None,
        collection_template: str | None = None,
        item_template: str | None = None,
        fallback_strategy: HrefLayoutStrategy | None = None,
    ):
        self.catalog_template = (
            LayoutTemplate(catalog_template) if catalog_template is not None else None
        )
        self.collection_template = (
            LayoutTemplate(collection_template)
            if collection_template is not None
            else None
        )
        self.item_template = (
            LayoutTemplate(item_template) if item_template is not None else None
        )

        if fallback_strategy is None:
            fallback_strategy = BestPracticesLayoutStrategy()
        self.fallback_strategy = fallback_strategy

    def get_catalog_href(self, cat: Catalog, parent_dir: str, is_root: bool) -> str:
        if is_root or self.catalog_template is None:
            return self.fallback_strategy.get_catalog_href(cat, parent_dir, is_root)
        else:
            template_path = self.catalog_template.substitute(cat)
            if not template_path.endswith(".json"):
                template_path = posixpath.join(template_path, cat.DEFAULT_FILE_NAME)

            return posixpath.join(parent_dir, template_path)

    def get_collection_href(
        self, col: Collection, parent_dir: str, is_root: bool
    ) -> str:
        if is_root or self.collection_template is None:
            return self.fallback_strategy.get_collection_href(col, parent_dir, is_root)
        else:
            template_path = self.collection_template.substitute(col)
            if not template_path.endswith(".json"):
                template_path = posixpath.join(template_path, col.DEFAULT_FILE_NAME)

            return posixpath.join(parent_dir, template_path)

    def get_item_href(self, item: Item, parent_dir: str) -> str:
        if self.item_template is None:
            return self.fallback_strategy.get_item_href(item, parent_dir)
        else:
            template_path = self.item_template.substitute(item)
            if not template_path.endswith(".json"):
                template_path = posixpath.join(template_path, f"{item.id}.json")

            return posixpath.join(parent_dir, template_path)


class BestPracticesLayoutStrategy(HrefLayoutStrategy):
    """Layout strategy that represents the catalog layout described
    in the :stac-spec:`STAC Best Practices documentation
    <best-practices.md>`

    For a root catalog or collection, this will use the filename 'catalog.json'
    or 'collection.json' to the given directory. For a non-root catalog or collection,
    the ID will be used as a subdirectory, e.g. ``${id}/catalog.json`` or
    ``${id}/collection.json``. For items, a subdirectory with a name of the item
    ID will be made, and the item ID will be used in the filename, i.e.
    ``${id}/${id}.json``

    All paths are appended to the parent directory.
    """

    def get_catalog_href(self, cat: Catalog, parent_dir: str, is_root: bool) -> str:
        if is_root:
            cat_root = parent_dir
        else:
            cat_root = posixpath.join(parent_dir, f"{cat.id}")

        return posixpath.join(cat_root, cat.DEFAULT_FILE_NAME)

    def get_collection_href(
        self, col: Collection, parent_dir: str, is_root: bool
    ) -> str:
        if is_root:
            col_root = parent_dir
        else:
            col_root = posixpath.join(parent_dir, f"{col.id}")

        return posixpath.join(col_root, col.DEFAULT_FILE_NAME)

    def get_item_href(self, item: Item, parent_dir: str) -> str:
        item_root = posixpath.join(parent_dir, f"{item.id}")

        return posixpath.join(item_root, f"{item.id}.json")


class AsIsLayoutStrategy(HrefLayoutStrategy):
    """Layout strategy that simply preserves the current href of all objects.

    If any object doesn't have a self href, a ValueError is raised.
    """

    def get_catalog_href(self, cat: Catalog, parent_dir: str, is_root: bool) -> str:
        href = cat.self_href
        if href is None:
            raise ValueError(
                f"Catalog is missing href, required for AsIsLayoutStrategy: {cat}"
            )
        else:
            return href

    def get_collection_href(
        self, col: Collection, parent_dir: str, is_root: bool
    ) -> str:
        href = col.self_href
        if href is None:
            raise ValueError(
                f"Collection is missing href, required for AsIsLayoutStrategy: {col}"
            )
        else:
            return href

    def get_item_href(self, item: Item, parent_dir: str) -> str:
        href = item.self_href
        if href is None:
            raise ValueError(
                f"Item is missing href, required for AsIsLayoutStrategy: {item}"
            )
        else:
            return href


class APILayoutStrategy(HrefLayoutStrategy):
    """Layout strategy that represents the STAC API endpoint layout described
    in the `STAC API Specifications
    <https://github.com/radiantearth/stac-api-spec/blob/v1.0.0/overview.md#endpoints>`_

    The URL of the root catalog will be the base URL of the API.
    Other catalogs will be listed underneath their parent catalog
    at ``./${id}``.
    Collections will be found underneath the root or their parent
    catalog at ``./collections/${id}``. Collection cannot be the root themselves.
    Items will be found underneath their collections at
    ``./collections/${collection}/items/${id}``.
    """

    def get_catalog_href(
        self, cat: Catalog | str, parent_dir: str, is_root: bool
    ) -> str:
        """
        Generate a catalog href based on the provided parameters.

        Parameters:
        cat (Catalog | str): The catalog object or its ID.
        parent_dir (str): The parent directory for the catalog.
        is_root (bool): A flag indicating whether the catalog is a root catalog.

        Returns:
        str: The generated catalog href.
        """

        if not isinstance(cat, str):
            cat = cat.id

        if is_root:
            cat_href = parent_dir
        else:
            cat_href = posixpath.join(parent_dir, cat)

        return cat_href

    def get_collections_href(self, parent_dir: str) -> str:
        """
        Generate a collections href based on the provided parent directory.

        Parameters:
        parent_dir (str): The parent directory for the collections.

        Returns:
        str: The generated collections href.
        """
        return posixpath.join(parent_dir, "collections")

    def get_collection_href(
        self, col: Collection | str, parent_dir: str, is_root: bool
    ) -> str:
        """
        Generate a collection href based on the provided parameters.

        Parameters:
        col (Collection | str): The collection object or its ID.
        parent_dir (str): The parent directory for the collection.
        is_root (bool): A flag indicating whether the collection is a root collection.

        Returns:
        str: The generated collection href.

        Raises:
        ValueError: If the collection is set as root.
        """

        if not isinstance(col, str):
            col = col.id

        if is_root:
            raise ValueError("Collections cannot be root")

        col_root = self.get_collections_href(parent_dir)

        return posixpath.join(col_root, col)

    def get_items_href(self, parent_dir: str) -> str:
        """
        Generate an items href based on the provided parent directory.

        Parameters:
        parent_dir (str): The parent directory for the items.

        Returns:
        str: The generated items href.
        """
        return posixpath.join(parent_dir, "items")

    def get_item_href(self, item: Item | str, parent_dir: str) -> str:
        """
        Generate an item href based on the provided parameters.

        Parameters:
        item (Item | str): The item object or its ID.
        parent_dir (str): The parent directory for the item.

        Returns:
        str: The generated item href.
        """
        if not isinstance(item, str):
            item = item.id

        item_root = self.get_items_href(parent_dir)

        return posixpath.join(item_root, item)

    def get_search_href(self, parent_dir: str) -> str:
        """
        Generate a search href based on the provided parent directory.

        Parameters:
        parent_dir (str): The parent directory for the search.

        Returns:
        str: The generated search href.
        """
        return posixpath.join(parent_dir, "search")

    def get_conformance_href(self, parent_dir: str) -> str:
        """
        Generate a conformance href based on the provided parent directory.

        Parameters:
        parent_dir (str): The parent directory for the conformance.

        Returns:
        str: The generated conformance href.
        """
        return posixpath.join(parent_dir, "conformance")

    def get_service_desc_href(self, parent_dir: str) -> str:
        """
        Generate an API service description href based on the provided parent directory.

        Parameters:
        parent_dir (str): The parent directory for the API.

        Returns:
        str: The generated API href.
        """
        return posixpath.join(parent_dir, "api")

    def get_service_doc_href(self, parent_dir: str) -> str:
        """
        Generate an API service documentation href based on
        the provided parent directory.

        Parameters:
        parent_dir (str): The parent directory for the API.

        Returns:
        str: The generated API href.
        """
        return posixpath.join(parent_dir, "api.html")



import json
import os
import tempfile
import unittest
from pathlib import Path

import pytest

import pystac
import pystac.errors
from pystac.stac_io import DefaultStacIO, DuplicateKeyReportingMixin, StacIO
from tests.utils import TestCases


def test_read_write_collection() -> None:
    collection = pystac.read_file(
        TestCases.get_path("data-files/collections/multi-extent.json")
    )
    with tempfile.TemporaryDirectory() as tmp_dir:
        dest_href = os.path.join(tmp_dir, "collection.json")
        pystac.write_file(collection, dest_href=dest_href)
        assert os.path.exists(dest_href), "File was not written."


def test_read_write_collection_with_file_protocol() -> None:
    collection = pystac.read_file(
        "file://" + TestCases.get_path("data-files/collections/multi-extent.json")
    )
    with tempfile.TemporaryDirectory() as tmp_dir:
        dest_href = os.path.join(tmp_dir, "collection.json")
        pystac.write_file(collection, dest_href="file://" + dest_href)
        assert os.path.exists(dest_href), "File was not written."


def test_read_item() -> None:
    item = pystac.read_file(TestCases.get_path("data-files/item/sample-item.json"))
    with tempfile.TemporaryDirectory() as tmp_dir:
        dest_href = os.path.join(tmp_dir, "item.json")
        pystac.write_file(item, dest_href=dest_href)
        assert os.path.exists(dest_href), "File was not written."


def test_read_write_catalog() -> None:
    catalog = pystac.read_file(
        TestCases.get_path("data-files/catalogs/test-case-1/catalog.json")
    )
    with tempfile.TemporaryDirectory() as tmp_dir:
        dest_href = os.path.join(tmp_dir, "catalog.json")
        pystac.write_file(catalog, dest_href=dest_href)
        assert os.path.exists(dest_href), "File was not written."


def test_read_item_collection_raises_exception() -> None:
    with pytest.raises(pystac.STACTypeError):
        _ = pystac.read_file(
            TestCases.get_path("data-files/item-collection/sample-item-collection.json")
        )


def test_read_item_dict() -> None:
    stac_io = StacIO.default()
    item_dict = stac_io.read_json(
        TestCases.get_path("data-files/item/sample-item.json")
    )
    item = pystac.read_dict(item_dict)
    assert isinstance(item, pystac.Item)


def test_read_collection_dict() -> None:
    stac_io = StacIO.default()
    collection_dict = stac_io.read_json(
        TestCases.get_path("data-files/collections/multi-extent.json")
    )
    collection = pystac.read_dict(collection_dict)
    assert isinstance(collection, pystac.Collection)


def test_read_catalog_dict() -> None:
    stac_io = StacIO.default()
    catalog_dict = stac_io.read_json(
        TestCases.get_path("data-files/catalogs/test-case-1/catalog.json")
    )
    catalog = pystac.read_dict(catalog_dict)
    assert isinstance(catalog, pystac.Catalog)


def test_read_from_stac_object() -> None:
    catalog = pystac.STACObject.from_file(
        TestCases.get_path("data-files/catalogs/test-case-1/catalog.json")
    )
    assert isinstance(catalog, pystac.Catalog)


def test_report_duplicate_keys() -> None:
    # Directly from dict
    class ReportingStacIO(DefaultStacIO, DuplicateKeyReportingMixin):
        pass

    stac_io = ReportingStacIO()
    test_json = """{
        "key": "value_1",
        "key": "value_2"
    }"""

    with pytest.raises(pystac.DuplicateObjectKeyError) as excinfo:
        stac_io.json_loads(test_json)
    assert str(excinfo.value) == 'Found duplicate object name "key"'

    # From file
    with tempfile.TemporaryDirectory() as tmp_dir:
        src_href = os.path.join(tmp_dir, "test.json")
        with open(src_href, "w") as dst:
            dst.write(test_json)

        with pytest.raises(pystac.DuplicateObjectKeyError) as excinfo:
            stac_io.read_json(src_href)
        assert str(excinfo.value), f'Found duplicate object name "key" in {src_href}'


@unittest.mock.patch("pystac.stac_io.urllib3.PoolManager.request")
def test_headers_stac_io(request_mock: unittest.mock.MagicMock) -> None:
    stac_io = DefaultStacIO(headers={"Authorization": "api-key fake-api-key-value"})

    catalog = pystac.Catalog("an-id", "a description").to_dict()
    # required until https://github.com/stac-utils/pystac/pull/896 is merged
    catalog["links"] = []
    request_mock.return_value.__enter__.return_value.read.return_value = json.dumps(
        catalog
    ).encode("utf-8")
    pystac.Catalog.from_file("https://example.com/catalog.json", stac_io=stac_io)

    headers = request_mock.call_args[1]["headers"]
    assert headers == {"User-Agent": f"pystac/{pystac.__version__}", **stac_io.headers}


@pytest.mark.vcr()
def test_retry_stac_io() -> None:
    # This doesn't test any retry behavior, but it does make sure that we can
    # still read objects.
    _ = pytest.importorskip("urllib3")
    from pystac.stac_io import RetryStacIO

    stac_io = RetryStacIO()
    _ = stac_io.read_stac_object("https://planetarycomputer.microsoft.com/api/stac/v1")


@pytest.mark.vcr()
def test_retry_stac_io_404() -> None:
    # This doesn't test any retry behavior, but it does make sure that we can
    # error when an object doesn't exist.
    _ = pytest.importorskip("urllib3")
    from pystac.stac_io import RetryStacIO

    stac_io = RetryStacIO()
    with pytest.raises(Exception):
        _ = stac_io.read_stac_object(
            "https://planetarycomputer.microsoft.com"
            "/api/stac/v1/collections/not-a-collection-id"
        )


def test_save_http_href_errors(tmp_path: Path) -> None:
    catalog = pystac.Catalog(id="test-catalog", description="")
    catalog.set_self_href("http://pystac.test/catalog.json")
    with pytest.raises(NotImplementedError):
        catalog.save_object()


@pytest.mark.vcr()
def test_urls_with_non_ascii_characters() -> None:
    from pystac.stac_io import HAS_URLLIB3

    url = "https://capella-open-data.s3.us-west-2.amazonaws.com/stac/capella-open-data-by-capital/capella-open-data-malé/collection.json"

    if HAS_URLLIB3:
        pystac.Collection.from_file(url)
    else:
        with pytest.raises(pystac.STACError):
            pystac.Collection.from_file(url)


@pytest.mark.vcr()
def test_proj_json_schema_is_readable() -> None:
    from pystac.stac_io import DefaultStacIO

    stac_io = DefaultStacIO()
    _ = stac_io.read_text_from_href(
        "https://proj.org/schemas/v0.7/projjson.schema.json"
    )



from __future__ import annotations

from collections import ChainMap
from copy import copy
from typing import TYPE_CHECKING, Any, cast

import pystac

if TYPE_CHECKING:
    from pystac.collection import Collection
    from pystac.stac_object import STACObject


def get_cache_key(stac_object: STACObject) -> tuple[str, bool]:
    """Produce a cache key for the given STAC object.

    If a self href is set, use that as the cache key.
    If not, use a key that combines this object's ID with
    its parents' IDs.

    Returns:
        Tuple[str, bool]: A tuple with the cache key as the first
            element and a boolean that is true if the cache key is
            the object's HREF as the second element.
    """
    href = stac_object.get_self_href()
    if href is not None:
        return href, True
    else:
        ids: list[str] = []
        obj: pystac.STACObject | None = stac_object
        while obj is not None:
            ids.append(obj.id)
            obj = obj.get_parent()
        return "/".join(ids), False


class ResolvedObjectCache:
    """This class tracks resolved objects tied to root catalogs.
    A STAC object is 'resolved' when it is a Python Object; a link
    to a STAC object such as a Catalog or Item is considered "unresolved"
    if it's target is pointed at an HREF of the object.

    Tracking resolved objects allows us to tie together the same instances
    when there are loops in the Graph of the STAC catalog (e.g. a LabelItem
    can link to a rel:source, and if that STAC Item exists in the same
    root catalog they should refer to the same Python object).

    Resolution tracking is important when copying STACs in-memory: In order
    for object links to refer to the copy of STAC Objects rather than their
    originals, we have to keep track of the resolved STAC Objects and replace
    them with their copies.

    Args:
        id_keys_to_objects : Existing cache of
            a key made up of the STACObject and it's parents IDs mapped
            to the cached STACObject.
        hrefs_to_objects : STAC Object HREFs matched to
            their cached object.
        ids_to_collections : Map of collection IDs
            to collections.
    """

    id_keys_to_objects: dict[str, STACObject]
    """Existing cache of a key made up of the STACObject and it's parents IDs mapped
    to the cached STACObject."""

    hrefs_to_objects: dict[str, STACObject]
    """STAC Object HREFs matched to their cached object."""

    ids_to_collections: dict[str, Collection]
    """Map of collection IDs to collections."""

    _collection_cache: ResolvedObjectCollectionCache | None

    def __init__(
        self,
        id_keys_to_objects: dict[str, STACObject] | None = None,
        hrefs_to_objects: dict[str, STACObject] | None = None,
        ids_to_collections: dict[str, Collection] | None = None,
    ):
        self.id_keys_to_objects = id_keys_to_objects or {}
        self.hrefs_to_objects = hrefs_to_objects or {}
        self.ids_to_collections = ids_to_collections or {}

        self._collection_cache = None

    def get_or_cache(self, obj: STACObject) -> STACObject:
        """Gets the STACObject that is the cached version of the given STACObject; or,
        if none exists, sets the cached object to the given object.

        Args:
            obj : The given object whose cache key will be checked
                against the cache.

        Returns:
            STACObject: Either the cached object that has the same cache key as the
            given object, or the given object.
        """
        key, is_href = get_cache_key(obj)
        if is_href:
            if key in self.hrefs_to_objects:
                return self.hrefs_to_objects[key]
            else:
                self.cache(obj)
                return obj
        else:
            if key in self.id_keys_to_objects:
                return self.id_keys_to_objects[key]
            else:
                self.cache(obj)
                return obj

    def get(self, obj: STACObject) -> STACObject | None:
        """Get the cached object that has the same cache key as the given object.

        Args:
            obj : The given object whose cache key will be checked against
                the cache.

        Returns:
            STACObject or None: Either the cached object that has the same cache key as
            the given object, or None
        """
        key, is_href = get_cache_key(obj)
        if is_href:
            return self.get_by_href(key)
        else:
            return self.id_keys_to_objects.get(key)

    def get_by_href(self, href: str) -> STACObject | None:
        """Gets the cached object at href.

        Args:
            href : The href to use as the key for the cached object.

        Returns:
            STACObject or None: Returns the STACObject if cached, otherwise None.
        """
        return self.hrefs_to_objects.get(href)

    def get_collection_by_id(self, id: str) -> Collection | None:
        """Retrieved a cached Collection by its ID.

        Args:
            id : The ID of the collection.

        Returns:
            Collection or None: Returns the collection if there is one cached
                with the given ID, otherwise None.
        """
        return self.ids_to_collections.get(id)

    def cache(self, obj: STACObject) -> None:
        """Set the given object into the cache.

        Args:
            obj : The object to cache
        """
        key, is_href = get_cache_key(obj)
        if is_href:
            self.hrefs_to_objects[key] = obj
        else:
            self.id_keys_to_objects[key] = obj

        if isinstance(obj, pystac.Collection):
            self.ids_to_collections[obj.id] = obj

    def remove(self, obj: STACObject) -> None:
        """Removes any cached object that matches the given object's cache key.

        Args:
            obj : The object to remove
        """
        key, is_href = get_cache_key(obj)

        if is_href:
            self.hrefs_to_objects.pop(key, None)
        else:
            self.id_keys_to_objects.pop(key, None)

        if obj.STAC_OBJECT_TYPE == pystac.STACObjectType.COLLECTION:
            self.id_keys_to_objects.pop(obj.id, None)

    def __contains__(self, obj: STACObject) -> bool:
        key, is_href = get_cache_key(obj)
        return (
            key in self.hrefs_to_objects if is_href else key in self.id_keys_to_objects
        )

    def contains_collection_id(self, collection_id: str) -> bool:
        """Returns True if there is a collection with given collection ID is cached."""
        return collection_id in self.ids_to_collections

    def as_collection_cache(self) -> CollectionCache:
        if self._collection_cache is None:
            self._collection_cache = ResolvedObjectCollectionCache(self)
        return self._collection_cache

    @staticmethod
    def merge(
        first: ResolvedObjectCache, second: ResolvedObjectCache
    ) -> ResolvedObjectCache:
        """Merges two ResolvedObjectCache.

        The merged cache will give preference to the first argument; that is, if there
        are cached keys that exist in both the first and second cache, the object cached
        in the first will be cached in the resulting merged ResolvedObjectCache.

        Args:
            first : The first cache to merge. This cache will be
                the preferred cache for objects in the case of ID conflicts.
            second : The second cache to merge.

        Returns:
            ResolvedObjectCache: The resulting merged cache.
        """
        merged = ResolvedObjectCache(
            id_keys_to_objects=dict(
                ChainMap(
                    copy(first.id_keys_to_objects), copy(second.id_keys_to_objects)
                )
            ),
            hrefs_to_objects=dict(
                ChainMap(copy(first.hrefs_to_objects), copy(second.hrefs_to_objects))
            ),
            ids_to_collections=dict(
                ChainMap(
                    copy(first.ids_to_collections), copy(second.ids_to_collections)
                )
            ),
        )

        merged._collection_cache = ResolvedObjectCollectionCache.merge(
            merged, first._collection_cache, second._collection_cache
        )

        return merged


class CollectionCache:
    """Cache of collections that can be used to avoid re-reading Collection
    JSON in :func:`pystac.serialization.merge_common_properties
    <pystac.serialization.common_properties.merge_common_properties>`.
    The CollectionCache will contain collections as either as dicts or PySTAC
    Collections, and will set Collection JSON that it reads in order to merge
    in common properties.
    """

    cached_ids: dict[str, Collection | dict[str, Any]]
    cached_hrefs: dict[str, Collection | dict[str, Any]]

    def __init__(
        self,
        cached_ids: dict[str, Collection | dict[str, Any]] | None = None,
        cached_hrefs: dict[str, Collection | dict[str, Any]] | None = None,
    ):
        self.cached_ids = cached_ids or {}
        self.cached_hrefs = cached_hrefs or {}

    def get_by_id(self, collection_id: str) -> Collection | dict[str, Any] | None:
        return self.cached_ids.get(collection_id)

    def get_by_href(self, href: str) -> Collection | dict[str, Any] | None:
        return self.cached_hrefs.get(href)

    def contains_id(self, collection_id: str) -> bool:
        return collection_id in self.cached_ids

    def cache(
        self,
        collection: Collection | dict[str, Any],
        href: str | None = None,
    ) -> None:
        """Caches a collection JSON."""
        if isinstance(collection, pystac.Collection):
            self.cached_ids[collection.id] = collection
        else:
            self.cached_ids[collection["id"]] = collection

        if href is not None:
            self.cached_hrefs[href] = collection


class ResolvedObjectCollectionCache(CollectionCache):
    resolved_object_cache: ResolvedObjectCache

    def __init__(
        self,
        resolved_object_cache: ResolvedObjectCache,
        cached_ids: dict[str, Collection | dict[str, Any]] | None = None,
        cached_hrefs: dict[str, Collection | dict[str, Any]] | None = None,
    ):
        super().__init__(cached_ids, cached_hrefs)
        self.resolved_object_cache = resolved_object_cache

    def get_by_id(self, collection_id: str) -> Collection | dict[str, Any] | None:
        result = self.resolved_object_cache.get_collection_by_id(collection_id)
        if result is None:
            return super().get_by_id(collection_id)
        else:
            return result

    def get_by_href(self, href: str) -> Collection | dict[str, Any] | None:
        result = self.resolved_object_cache.get_by_href(href)
        if result is None:
            return super().get_by_href(href)
        else:
            return cast(pystac.Collection, result)

    def contains_id(self, collection_id: str) -> bool:
        return self.resolved_object_cache.contains_collection_id(
            collection_id
        ) or super().contains_id(collection_id)

    def cache(
        self,
        collection: Collection | dict[str, Any],
        href: str | None = None,
    ) -> None:
        super().cache(collection, href)

    @staticmethod
    def merge(
        resolved_object_cache: ResolvedObjectCache,
        first: ResolvedObjectCollectionCache | None,
        second: ResolvedObjectCollectionCache | None,
    ) -> ResolvedObjectCollectionCache:
        first_cached_ids = {}
        if first is not None:
            first_cached_ids = copy(first.cached_ids)

        second_cached_ids = {}
        if second is not None:
            second_cached_ids = copy(second.cached_ids)

        first_cached_hrefs = {}
        if first is not None:
            first_cached_hrefs = copy(first.cached_hrefs)

        second_cached_hrefs = {}
        if second is not None:
            second_cached_hrefs = copy(second.cached_hrefs)

        return ResolvedObjectCollectionCache(
            resolved_object_cache,
            cached_ids=dict(ChainMap(first_cached_ids, second_cached_ids)),
            cached_hrefs=dict(ChainMap(first_cached_hrefs, second_cached_hrefs)),
        )



"""Implements the :stac-ext:`Versioning Indicators Extension <version>`."""

from __future__ import annotations

from collections.abc import Generator
from contextlib import contextmanager
from typing import (
    Any,
    Generic,
    Literal,
    TypeVar,
    cast,
)

from pystac import (
    Asset,
    Catalog,
    Collection,
    ExtensionTypeError,
    Item,
    ItemAssetDefinition,
    Link,
    MediaType,
    STACObject,
    STACObjectType,
)
from pystac.errors import DeprecatedWarning
from pystac.extensions.base import ExtensionManagementMixin, PropertiesExtension
from pystac.extensions.hooks import ExtensionHooks
from pystac.utils import StringEnum, map_opt

#: Generalized version of :class:`~pystac.Catalog`, :class:`~pystac.Collection`,
#: :class:`~pystac.Item`, :class:`~pystac.Asset` or :class:`~pystac.ItemAssetDefinition`
T = TypeVar("T", Catalog, Collection, Item, Asset, ItemAssetDefinition)

#: Generalized version of :class:`~pystac.Catalog`, :class:`~pystac.Collection`, or
#: :class:`~pystac.Item`
U = TypeVar("U", Catalog, Collection, Item)

SCHEMA_URI = "https://stac-extensions.github.io/version/v1.2.0/schema.json"

# STAC fields - These are unusual for an extension in that they do not have
# a prefix.  e.g. nothing like "ver:"
VERSION: str = "version"
DEPRECATED: str = "deprecated"
EXPERIMENTAL: str = "experimental"


class VersionRelType(StringEnum):
    """A list of rel types defined in the Version Extension.

    See the :stac-ext:`Version Extension Relation types
    <version#relation-types>` documentation
    for details."""

    LATEST = "latest-version"
    """Indicates a link pointing to a resource containing the latest version."""

    PREDECESSOR = "predecessor-version"
    """Indicates a link pointing to a resource containing the predecessor version in the
    version history."""

    SUCCESSOR = "successor-version"
    """Indicates a link pointing to a resource containing the successor version in the
    version history."""

    HISTORY = "version-history"
    """This link points to a version history or changelog.
    
    This can be for example a Markdown file with the corresponding media type or
    a STAC Catalog or Collection.
    """


class BaseVersionExtension(
    Generic[T],
    PropertiesExtension,
    ExtensionManagementMixin[Catalog | Collection | Item],
):
    """A base extension that just gets and sets attributes, not links.

    Used for Assets and AssetDefinitions.
    """

    name: Literal["version"] = "version"

    def apply_base(
        self,
        version: str | None = None,
        deprecated: bool | None = None,
        experimental: bool | None = None,
    ) -> None:
        """Applies attributes to this extension object."""
        self.version = version
        self.deprecated = deprecated
        self.experimental = experimental

    @property
    def version(self) -> str | None:
        """Get or sets a version string of this object."""
        return self._get_property(VERSION, str)

    @version.setter
    def version(self, v: str | None) -> None:
        self._set_property(VERSION, v, pop_if_none=True)

    @property
    def deprecated(self) -> bool | None:
        """Get or sets whether this object is deprecated.

        A value of ``True`` specifies that the object is deprecated with the
        potential to be removed. It should be transitioned out of usage as soon
        as possible and users should refrain from using it in new projects. A
        link with relation type ``latest-version`` SHOULD be added to the links
        and MUST refer to the resource that can be used instead.

        NOTE:
            A :class:`pystac.DeprecatedWarning` is issued if the ``deprecated``
            property is ``True`` when deserializing a dictionary to an object.
            The :meth:`~pystac.extensions.version.ignore_deprecated` context
            manager is provided as a convenience to suppress these warnings:

            >>> with ignore_deprecated():
            ...    deprecated_item = pystac.Item.from_dict(deprecated_item_dict)
        """
        return self._get_property(DEPRECATED, bool)

    @deprecated.setter
    def deprecated(self, v: bool | None) -> None:
        self._set_property(DEPRECATED, v, pop_if_none=True)

    @property
    def experimental(self) -> bool | None:
        """Get and set whether this object is experimental.

        Specifies that the context this field is used in (e.g. Asset or
        Collection) is experimental with the potential to break or be unstable.
        """
        return self._get_property(EXPERIMENTAL, bool)

    @experimental.setter
    def experimental(self, v: bool | None) -> None:
        self._set_property(EXPERIMENTAL, v, pop_if_none=True)

    @classmethod
    def get_schema_uri(cls) -> str:
        return SCHEMA_URI

    @classmethod
    def ext(cls, obj: T, add_if_missing: bool = False) -> BaseVersionExtension[T]:
        """Extends the given STAC Object with properties from the :stac-ext:`Versioning
        Indicators Extension <version>`.

        Raises:

            pystac.ExtensionTypeError : If an invalid object type is passed.
        """
        if isinstance(obj, Collection):
            cls.ensure_has_extension(obj, add_if_missing)
            return cast(BaseVersionExtension[T], CollectionVersionExtension(obj))
        elif isinstance(obj, Catalog):
            cls.ensure_has_extension(obj, add_if_missing)
            return cast(BaseVersionExtension[T], CatalogVersionExtension(obj))
        elif isinstance(obj, Item):
            cls.ensure_has_extension(obj, add_if_missing)
            return cast(BaseVersionExtension[T], ItemVersionExtension(obj))
        elif isinstance(obj, Asset):
            cls.ensure_owner_has_extension(obj, add_if_missing)
            return cast(BaseVersionExtension[T], AssetVersionExtension(obj))
        else:
            raise ExtensionTypeError(cls._ext_error_message(obj))


class VersionExtension(
    Generic[U],
    BaseVersionExtension[U],
):
    """An abstract class that can be used to extend the properties of an
    :class:`~pystac.Item`, :class:`~pystac.Collection`, or
    :class:`~pystac.Catalog` with properties from the :stac-ext:`Versioning
    Indicators Extension <version>`. This class is generic over the type of STAC
    Object to be extended.

    To create a concrete instance of :class:`VersionExtension`, use the
    :meth:`VersionExtension.ext` method. For example:

    .. code-block:: python

       >>> item: pystac.Item = ...
       >>> version_ext = VersionExtension.ext(item)
    """

    obj: STACObject

    def __init__(self, obj: STACObject) -> None:
        self.obj = obj

    def apply(
        self,
        version: str | None = None,
        deprecated: bool | None = None,
        experimental: bool | None = None,
        latest: U | None = None,
        predecessor: U | None = None,
        successor: U | None = None,
    ) -> None:
        """Applies version extension properties to the extended :class:`~pystac.Item` or
        :class:`~pystac.Collection`.

        Args:
            version : The version string for the item.
            deprecated : Optional flag set to ``True`` if an Item is deprecated with the
                potential to be removed.  Defaults to ``False`` if not present.
            latest : Item representing the latest (e.g., current) version.
            predecessor : Item representing the resource containing the predecessor
                version in the version history.
            successor : Item representing the resource containing the successor version
                in the version history.
        """
        self.apply_base(version, deprecated, experimental)
        if latest:
            self.latest = latest
        if predecessor:
            self.predecessor = predecessor
        if successor:
            self.successor = successor

    @property
    def latest(self) -> U | None:
        """Gets or sets the :class:`~pystac.Link` to the :class:`~pystac.Item`
        representing the most recent version.
        """
        return map_opt(
            lambda x: cast(U, x),
            next(iter(self.obj.get_stac_objects(VersionRelType.LATEST)), None),
        )

    @latest.setter
    def latest(self, value: U | None) -> None:
        self.obj.clear_links(VersionRelType.LATEST)
        if value is not None:
            self.obj.add_link(Link(VersionRelType.LATEST, value, MediaType.JSON))

    @property
    def predecessor(self) -> U | None:
        """Gets or sets the :class:`~pystac.Link` to the :class:`~pystac.Item`
        representing the resource containing the predecessor version in the version
        history.
        """
        return map_opt(
            lambda x: cast(U, x),
            next(iter(self.obj.get_stac_objects(VersionRelType.PREDECESSOR)), None),
        )

    @predecessor.setter
    def predecessor(self, value: U | None) -> None:
        self.obj.clear_links(VersionRelType.PREDECESSOR)
        if value is not None:
            self.obj.add_link(
                Link(
                    VersionRelType.PREDECESSOR,
                    value,
                    MediaType.JSON,
                )
            )

    @property
    def successor(self) -> U | None:
        """Gets or sets the :class:`~pystac.Link` to the :class:`~pystac.Item`
        representing the resource containing the successor version in the version
        history.
        """
        return map_opt(
            lambda x: cast(U, x),
            next(iter(self.obj.get_stac_objects(VersionRelType.SUCCESSOR)), None),
        )

    @successor.setter
    def successor(self, value: U | None) -> None:
        self.obj.clear_links(VersionRelType.SUCCESSOR)
        if value is not None:
            self.obj.add_link(Link(VersionRelType.SUCCESSOR, value, MediaType.JSON))

    @classmethod
    def ext(cls, obj: U, add_if_missing: bool = False) -> VersionExtension[U]:
        """Extends the given STAC Object with properties from the :stac-ext:`Versioning
        Indicators Extension <version>`.

        This extension can be applied to instances of :class:`~pystac.Item` or
        :class:`~pystac.Collection`.

        Raises:

            pystac.ExtensionTypeError : If an invalid object type is passed.
        """
        if isinstance(obj, Collection):
            cls.ensure_has_extension(obj, add_if_missing)
            return cast(VersionExtension[U], CollectionVersionExtension(obj))
        elif isinstance(obj, Catalog):
            cls.ensure_has_extension(obj, add_if_missing)
            return cast(VersionExtension[U], CatalogVersionExtension(obj))
        elif isinstance(obj, Item):
            cls.ensure_has_extension(obj, add_if_missing)
            return cast(VersionExtension[U], ItemVersionExtension(obj))
        else:
            raise ExtensionTypeError(cls._ext_error_message(obj))


class CatalogVersionExtension(VersionExtension[Catalog]):
    """A concrete implementation of :class:`VersionExtension` on a
    :class:`~pystac.Catalog` that extends the properties of the Catalog to
    include properties defined in the :stac-ext:`Versioning Indicators Extension
    <version>`.

    This class should generally not be instantiated directly. Instead, call
    :meth:`VersionExtension.ext` on an :class:`~pystac.Catalog` to extend it.
    """

    catalog: Catalog
    links: list[Link]
    properties: dict[str, Any]

    def __init__(self, catalog: Catalog):
        self.catalog = catalog
        self.properties = catalog.extra_fields
        self.links = catalog.links
        super().__init__(self.catalog)

    def __repr__(self) -> str:
        return f"<CatalogVersionExtension Item id={self.catalog.id}>"


class CollectionVersionExtension(VersionExtension[Collection]):
    """A concrete implementation of :class:`VersionExtension` on a
    :class:`~pystac.Collection` that extends the properties of the Collection to
    include properties defined in the :stac-ext:`Versioning Indicators Extension
    <version>`.

    This class should generally not be instantiated directly. Instead, call
    :meth:`VersionExtension.ext` on an :class:`~pystac.Collection` to extend it.
    """

    collection: Collection
    links: list[Link]
    properties: dict[str, Any]

    def __init__(self, collection: Collection):
        self.collection = collection
        self.properties = collection.extra_fields
        self.links = collection.links
        super().__init__(self.collection)

    def __repr__(self) -> str:
        return f"<CollectionVersionExtension Item id={self.collection.id}>"


class ItemVersionExtension(VersionExtension[Item]):
    """A concrete implementation of :class:`VersionExtension` on an
    :class:`~pystac.Item` that extends the properties of the Item to include properties
    defined in the :stac-ext:`Versioning Indicators Extension <version>`.

    This class should generally not be instantiated directly. Instead, call
    :meth:`VersionExtension.ext` on an :class:`~pystac.Item` to extend it.
    """

    item: Item
    links: list[Link]
    properties: dict[str, Any]

    def __init__(self, item: Item):
        self.item = item
        self.properties = item.properties
        self.links = item.links
        super().__init__(self.item)

    def __repr__(self) -> str:
        return f"<ItemVersionExtension Item id={self.item.id}>"


class AssetVersionExtension(BaseVersionExtension[Asset]):
    """A concrete implementation of :class:`VersionExtension` on an
    :class:`~pystac.Asset` that extends the properties of the Asset to include
    properties defined in the :stac-ext:`Versioning Indicators Extension
    <version>`.

    This class should generally not be instantiated directly. Instead, call
    :meth:`VersionExtension.ext` on an :class:`~pystac.Asset` to extend it.
    """

    asset: Asset
    properties: dict[str, Any]

    def __init__(self, asset: Asset):
        self.asset = asset
        self.properties = asset.extra_fields

    def __repr__(self) -> str:
        return f"<AssetVersionExtension Asset href={self.asset.href}>"


class ItemAssetsViewExtension(BaseVersionExtension[ItemAssetDefinition]):
    properties: dict[str, Any]

    def __init__(self, item_asset: ItemAssetDefinition):
        self.properties = item_asset.properties


class VersionExtensionHooks(ExtensionHooks):
    schema_uri = SCHEMA_URI
    prev_extension_ids = {
        "version",
        "https://stac-extensions.github.io/version/v1.0.0/schema.json",
        "https://stac-extensions.github.io/version/v1.1.0/schema.json",
    }
    stac_object_types = {
        STACObjectType.COLLECTION,
        STACObjectType.ITEM,
        STACObjectType.CATALOG,
    }

    def get_object_links(self, obj: STACObject) -> list[str] | None:
        if isinstance(obj, Collection) or isinstance(obj, Item):
            return [
                VersionRelType.LATEST,
                VersionRelType.PREDECESSOR,
                VersionRelType.SUCCESSOR,
            ]
        return None


VERSION_EXTENSION_HOOKS: ExtensionHooks = VersionExtensionHooks()


@contextmanager
def ignore_deprecated() -> Generator[None]:
    """Context manager for suppressing the :class:`pystac.DeprecatedWarning`
    when creating a deprecated :class:`~pystac.Item` or :class:`~pystac.Collection`
    from a dictionary."""
    import warnings

    with warnings.catch_warnings():
        warnings.simplefilter("ignore", category=DeprecatedWarning)
        yield
